/****************************************************************************************************************************************************
                                                      EFIS_EMS_Teensy_AvionicsDuino_CarloM_V1_0
******************************************************************************************************************************************************
    
  EFIS_EMS_Teensy_AvionicsDuino_CarloM_V1_0 is free software    
  MIT License (MIT)
  
  Copyright (c) 2025 AvionicsDuino
  Contributors : Benjamin Frémond - Carlo Marcheggiani - Gabriel Consigny - Michel Riazuelo
  https://avionicsduino.com/index.php/en/efis-2/

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
  of the Software, and to permit persons to whom the Software is furnished to do so,
  subject to the following conditions:
  
  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.
  
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
 *****************************************************************************************************************************/

// --------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                Connexions physiques avec la carte Teensy 4.1
// --------------------------------------------------------------------------------------------------------------------------------------------------------------
// Ecran : module TFT LCD RIVERDI (ref. RVT43HLTFWN00), avec contrôleur RA8875 Adafruit(ref. Product ID: 1590)
// ---------------- RA8875 --------------------- Teensy 4.1 ------------------------
//                   GND           --------->      GND
//                   VIN           --------->   + 5 Volts
//                    CS           --------->   Pin 0 CS1
//                   MISO          --------->   Pin 1 MISO1
//                   MOSI          --------->   Pin 26 MOSI1
//                   SCK           --------->   Pin 27 SCK1
//             Tout le reste       --------->   Non connecté
//---------------------------------------------------------------------------
// Encodeur rotatif optique GrayHill :
// ---------------GrayHill------------------------Teensy 4.1------------------------
//                  1    ---------------------->   GND
//                  2    ---------------------->   GND
//                6 Vcc  ---------------------->   3.3 v
//                3 SW   ---------------------->   Pin 35 Bouton central
//                 4 B   ---------------------->   Pin 37 Phase B
//                 5 A   ---------------------->   Pin 36 Phase A
//---------------------------------------------------------------------------
// Capteur de température intérieure (oneWire DS18B20) :
// -------------DS18B20------------------------Teensy 4.1------------------------
//                GND     ------------------->   GND
//             Alim +3.3v ------------------->   + 3.3 volts (courant : 7 mA)
//              Signal   -------------------->   pin 2 + résistance pull up de 4,7k vers 3.3 volts
//---------------------------------------------------------------------------
// Capteurs de pression AMS5915 (1500A pour l'altitude, 050D pour la vitesse air, il sont exploités par une adaptation de cette library : https://github.com/bolderflight/ams5915-arduino
// -------------1500A------------------------Teensy 4.1------------------------
//                GND     ------------------->   GND
//            Alim +3.3v  ------------------->   + 3.3 volts (courant : 4 mA)
//                SDA   --------------------->   SDA (pin 18) + résistance pull up de 4,7k vers 3.3 volts
//                SCL   --------------------->   SCL (pin 19) + résistance pull up de 4,7k vers 3.3 volts
//           Adresse : 0x28
//---------------------------------------------------------------------------
// Capteur de pression absolue pour la PA : Adafruit MPRLS
//----------------MPRLS ------------------------- Teensy 4.1 -----------------------
//                 Vin --------------------------> Vin
//                 GND --------------------------> GND
//                 SCL --------------------------> Pin 24 (SCL2)
//                 SDA --------------------------> Pin 25 (SDA2)
//            Autres broches --------------------> Non connectées
//---------------------------------------------------------------------------
// -------------0050D------------------------Teensy 4.1------------------------
//                GND     ------------------->   GND
//            Alim +3.3v   ------------------>   + 3.3 volts (courant : 4 mA)
//                SDA   --------------------->   SDA1 (pin 17) + résistance pull up de 4,7k vers 3.3 volts
//                SCL   --------------------->   SCL1 (pin 16) + résistance pull up de 4,7k vers 3.3 volts
//           Adresse : 0x28
//---------------------------------------------------------------------------
// Le transceiver CAN MCP 2562 EP (CAN 2.0) :
//-------------- MCP 2562 ------------------------ Teensy 4.1 ------------------------
//             Pin1 TXD ------------------------> Pin 31 CTX3
//             Pin2 GND ------------------------> GND
//             Pin3 VDD ------------------------> + 5V
//             Pin4 RXD ------------------------> Pin 30 CRX3
//             Pin5 Vio ------------------------> + 3.3V
//             Pin6 CAN LOW --------------------> sortie CanBus
//             Pin7 CAN HIGH -------------------> Sortie CanBus
//             Pin8 Standby --------------------> GND

//---------------------------------------------------------------------------
// Le module AHRS Avionics Duino est connecté au CAN Bus
// --------------------------------------------------------------------------

//----------------------------------------------------------------------------------------------------------------
// L'heure locale est mémorisée dans la SRTC de la carte Teensy 4.1 par une pile lithium 3v CR2032 entre GND et VBAT.
//----------------------------------------------------------------------------------------------------------------

//-----------------------------------------------------------------------------------------------------------------
// Une carte micro SD formatée ExFAT et contenant le fichier MTBLANC.out doit être insérée dans le lecteur de carte
//-----------------------------------------------------------------------------------------------------------------

// --------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                  Inclusions des bibliothèques et fichiers externes
// --------------------------------------------------------------------------------------------------------------------------------------------------------------

#include <EEPROM.h>
#include <SPI.h>
#include <RA8875.h>
#include <OneWire.h>
#include <TimeLib.h>
#include <SD.h>
#include <QuadEncoder.h>
#include <Adafruit_GFX.h>  // Pour les fontes, il faut télécharger cette library https://github.com/mjs513/ILI9341_fonts et l'installer dans le dossier des libraries Teensy.
#include <font_Arial.h>
#include <font_ArialBold.h>
#include <TeensyTimerTool.h>
using namespace TeensyTimerTool;
#include "Adafruit_MPRLS.h"
#include "HorizArt.h"
#include "HorizArtParamGen.h"
#include "AMS5915_simplified.h"
#include <FlexCAN_T4.h>
#include "EMS.h"
#include "FUEL.h"
#include <elapsedMillis.h>

// --------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                             Création des objets
// --------------------------------------------------------------------------------------------------------------------------------------------------------------
RA8875 tft = RA8875(RA8875_CS1, 255, RA8875_MOSI1, RA8875_SCLK1, RA8875_MISO1);  // Crée l'objet RA8875 "tft", de classe RA8875, initialisé dans le setup
OneWire ds(2);
AMS5915_simplified AMS5915_050D(Wire, 0x28, AMS5915_simplified::AMS5915_0050_D);
AMS5915_simplified AMS5915_1500A(Wire1, 0x28, AMS5915_simplified::AMS5915_1500_A);
HorizArt Horizon(&tft);
QuadEncoder encoder(1, 36, 37, 1);  // Canal 1, Phase A (pin36), PhaseB(pin37), pullups nécessaire avec l'encodeur GrayHill(1.
FlexCAN_T4<CAN3, RX_SIZE_256, TX_SIZE_16> CAN_Module_EFIS;
File file;
PeriodicTimer TimerClock(TCK);
#define RESET_PIN -1  // no hard-reset pin
#define EOC_PIN -1    // no end-of-conversion pin
Adafruit_MPRLS mpr = Adafruit_MPRLS(RESET_PIN, EOC_PIN);

// --------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                             Déclarations des variables et constantes globales
// --------------------------------------------------------------------------------------------------------------------------------------------------------------

// ******************************************* Déclaration du typedef d'une structure, puis d'un tableau d'éléments de cette structure, définissant l'arborescence des menus ********************************************************
/* La structure des options de menus et l'algorithme de traitement de ces options ont été conçus de telle façon qu'il soit très simple de faire évoluer ces menus,
   de modifier, de supprimer ou de rajouter des options.
   Voir le schéma du système de menus : https://i.imgur.com/oD18FKH.jpg.
   Voir le schéma de l'algorithme : https://i.imgur.com/eA0QNdb.jpg
   Voir le rôle des index : https://i.imgur.com/c1LsPR6.jpg
*/

typedef struct
{
  const char* menuItemLabel;
  uint32_t menuItemNumber;
  char menuItemAction;
  uint8_t numberOfSisterOptions;
  uint8_t indexOfParentOption;
} OptionMenu;

// Alarm structure
enum  almclass
{
  NOP,
  ALARM,
  WARNING,
  ADVISOR
};

typedef struct
{
  char msg[50];
	enum almclass errclass;
  uint32_t priority;
  uint8_t number;
} alarm;

OptionMenu menu[] = {
  { "FLIGHT", 10, 'S' },
  { " BACK", 100, 'B' },
  { "  QNH", 101, 'P' },
  { "  QFE", 102, 'P' },
  { "ResetG", 103, 'A' },
  { "  QUIT", 104, 'Q' },

  { "   GEN", 11, 'S' },

  { " BACK", 110, 'B' },
  { " Units", 111, 'S' },

  { " BACK", 1110, 'B' },
  { " Km/h", 1111, 'A' },
  { "Knots", 1112, 'A' },
  { "  mph", 1113, 'A' },
  { "  hPa", 1114, 'A' },
  { "  inHg", 1115, 'A' },
  { "Celsius", 1116, 'A' },
  { "Farenht", 1117, 'A' },
  { "  QUIT", 1118, 'Q' },

  { "MgDev", 112, 'P' },
  { "PrsCor", 113, 'P' },
  { "Backlit", 114, 'P' },
  { "Loc-utc", 115, 'P' },
  { "  QUIT", 116, 'Q' },

  { "Timers", 12, 'S' },

  { " BACK", 120, 'B' },
  { " Start", 121, 'A' },
  { " T. Up", 122, 'A' },
  { "T.Down", 123, 'P' },
  { "  QUIT", 124, 'Q' },

  { "  EIS ", 13, 'S' },

  { "  BACK", 130, 'B' },
  { "C.Flow", 131, 'P' },
  { "FuelAv", 132, 'P' },
  { "  QUIT", 133, 'Q' },

  { "  QUIT", 14, 'Q' },

  { " ", 1000000, '\0' }
};

// ********************************************************************* Autres variables Menus **************************************************************************
bool Menu_open = false;
bool SetVal = false;
uint8_t indexCurrentMenuOption = 0;
int16_t* ptrGen_int16_t = NULL;
float* ptrGen_float = NULL;
uint16_t menuXCoordinate = 0, menuYCoordinate = 254;
uint8_t boxHeight = 18, boxWidth = 53;

// ******************************************************************** Variables Encodeur ***********************************************************************
int EncoderPositionValue;
int EncoderPositionValueOld = 0;
int EncoderPositionValueOldSetVal = 0;
byte Button_Pin = 35;
const uint16_t debounceTime = 500;
volatile unsigned long lastButtonPress;
bool Button_Click = false;

// ************************************************************** Variables EFIS - horizon artificiel ***************************************************************
int16_t QNH, QFE;                    // QNH et QFE sont exprimés en centièmes de pouce de mercure, donc atmosphère ISA -> 2992 centièmes de inHg (soit 1013.25 hPa)
uint8_t displayBaroSettingUnit = 0;  // 0 -> hPa    1 -> In Hg
uint8_t displayTemperatureUnit = 0;  // 0 -> °C     1 -> °F
int16_t pressureCorrection;          // Correction d'un offset de pression absolue, variable, exprimée en déciPa. Peut être ajustée dans les menus pour corriger l'erreur altimétrique.
float altitudeCorrection = 0.995;    // Idem, mais pour "rattraper" une erreur proportionnelle à l'altidude.
float pressure = 101300;
float nonFilteredPressure;
float previousFilteredPressureValue;
float differentialPressure = 0;
float nonFilteredDifferentialPressure;
float previousFilteredDiffPressureValue = 0;
float qnhAltitude = 0;
float qfeAltitude = 0;
float pressureAltitude = 0;
float densityAltitude = 0.0;
float iat = 5.0;
volatile float oat = 5.0;
volatile float relativeHumidity = 60.0;
volatile float magneticHeading = 0.0;
float previousFilteredMagneticHeadingValue;
uint16_t intMagneticHeading;
float dewPoint = 0;
int16_t indicatedAirSpeed, trueAirSpeed;
int16_t nonFilteredIndicatedAirSpeed, nonFilteredTrueAirSpeed;
int16_t ballOffset = 0;
float altitudeVario, altitudeVarioOld;
float baroVario;
uint32_t baroVarioIntegrationStartTime;
uint16_t baroVarioIntegrationTimeOut = 250;
float previousFilteredBaroVarioValue;
float accG, accGmax = 1.0, accGmin = 1.0;
float previousFilteredAccGValue = 1;
float speedConversionFactor = 1.0;
char strSpeedUnit[] = "Km/h";
int16_t magDev = 0;
float windSpeed, windDirection;
float manifoldPressure = 17.0, nonFilteredManifoldPressure = 17.0, previousFilteredManifoldPressureValue = 17.0;  // mBar (0.5 HgInch)

// ******************************************************* Déclaration de 2 variables pour la gestion de l'EEPROM **********************************************
#define EFISeepromAdresse 22
uint16_t OffsetEFISeepromAdresse = 0;

// ************************************************************ Variables utilisées pour le chronométrage de la boucle principale **********************************************************
int loopCounter = 0;
int numLoopIterationDuration = 0;
float numLoopIteration = 1000.0;
float loopDuration;
unsigned long loopDurationStartTime;

// ****************************************************************** Variables utilisées pour l'exploitation des paramètres reçus de l'AHRS **************************************************
float roll = 0, pitch = 0, accY = 0, AccZ, Vz, trk, satInView;  // pour stocker les variables float de l'AHRS
float previousFilteredRollValue = 0, previousFilteredPitchValue = 0, previousFilteredAccYValue = 0, previousFilteredVzValue = 0, previousFilteredTrkValue;
int16_t roundedIntVz;  // pour stocker les variables entières de l'AHRS

// ************************************************************************* Variables GNSS *******************************************************************************************************************
float altitudeGNSS, altitudeGNSS_Old;
float groundSpeedGNSS, groundSpeedGNSS_Old;
uint16_t yearGNSS;
uint8_t monthGNSS;
uint8_t dayGNSS;
uint8_t hourGNSS;
uint8_t minuteGNSS;
uint8_t secondGNSS;
int32_t nanoSecGNSS;
uint32_t timeAccuracyGNSS;
bool fixOK = false;
bool flagTimeSet = false;

// ******************************************************* Variables et constantes liées à l'horloge et à la gestion de l'heure *****************************************************************************
#define Chy 33
#define Chx 131
#define Rh 28
byte hourHandLength, minuteHandLength, secondHandLength;
time_t zt;
time_t t_chrono1;
time_t t_chrono2;
time_t t_timer;
bool flag_chrono1 = false;
bool flag_chrono2 = false;
bool flagTimer = false;
int16_t timerDuration = 0;
bool flagDisplayTime = true;
uint16_t hs, ms;
uint8_t ss;
#define numberOfSecondsIn12h 43200
uint16_t angleH, angleM, angleS;
uint16_t angleHold = 0;
uint16_t angleMold = 0;
uint16_t angleSold = 0;
int16_t localTimeCorrection = 1;

// ************************************************************** Variables nécessaires au fonctionnement de la sonde de température intérieure********************************************************************************************
byte dataDS18B20[12];
byte addrDS18B20[8];
float temperatureDS18B20;
uint32_t DS18B20AcquisitionStartTime = 0;
bool flagDS18B20AcquisitionDelayElapsed = true;

// ************************************************************* Controllo ricezione messaggi dai devices CAN *******************************
uint16_t NoRDACmsg = 0;  // Missing Messages counter received from RDAC
uint16_t NoAHRSmsg = 0;  // Missing Messages counter received from AHRS
uint16_t NoFUELmsg = 0;  // Missing Messages counter received from Fuel System
uint16_t NoRELAYmsg = 0;  // Missing Messages counter received from Relay Unit
bool NoRDACflag = false;
bool NoAHRSflag = false;
bool NoFUELflag = false;
bool NoRELAYflag = false;

// ************************************************************* Variabili RDAC *************************************************************
uint16_t fuelflowcount = 0;                  // Pulses within last 4 seconds ID 1, 0-1
uint16_t fuelflowratio = 65535;              // Pulses duty cycle x10 ID 1, 2-3 (400 = 60.0/40.0% ratio) No pulse detected = 65535
int16_t tc[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };  // Thermocuples ID 2, 0-1, 2-3, 4-5, 6-7; ID 3, 0-1, 2-3, 4-5, 6-7
uint16_t oiltraw = 0;                        // Oil temp temperature ADC raw word ID 5, 0-1
uint16_t oilpraw = 0;                        // Oil press temperature ADC raw word ID 5, 2-3
uint16_t aux1raw = 0;                        // Aux1 ADC raw word ID 5, 4-5
uint16_t aux2raw = 0;                        // Aux2 ADC raw word ID 5, 6-7
uint16_t fuelpraw = 0;                       // Fuel press ADC raw word ID 6, 0-1
uint16_t coolraw = 0;                        // Coolant temp ADC raw word ID 6, 2-3
uint16_t fuellev1raw = 0;                    // Fuel level1 ADC raw word ID 6, 4-5
uint16_t fuellev2raw = 0;                    // Fuel level2 ADC raw word ID 6, 6-7
int16_t coldjtemp = -100;                    // Cold junction temp for TC compensation ID 7, 0-1 (temperature = TC + coldjtemp)
uint16_t RDACvolt = 0;                       // RDAC power voltage ID 7, 2-3 (volt=round(RDACvolt/5.73758))
uint16_t rpm = 0;                            // RPM from CAN Bus ID 8, 0-1
uint16_t curraw = 0;                         // MAP from CAN Bus ID 8, 6-7

// ************************************************************** Variabili EMS *******************************************************************
uint32_t alarmflag = 0;                       // Flags di allarme
uint32_t checked = 0;                         // Flags di allarme verificati
uint32_t fuelpulse = 0;                       // Total flowmeter pulses
uint16_t fuelburnt = 0;                       // Fuel burnt
int16_t availablefuel = 76;                   // Loaded fuel into the tanks (set in menu)
int16_t flowDiv = 17966;                      // Division factor for fluxometer (set in menu)
uint16_t lefttankraw = 0;                     // Fuel in left tank raw from CAN bus
uint16_t righttankraw = 0;                    // Fuel in right tank raw from CAN bus
uint16_t auxtankraw = 0;                      // Fuel in aux tank raw from CAN bus
uint16_t fuelalarmflags = 0;                  // Alarm flags from Fuel Measurement System
uint16_t rpmold = 0;                          // update or not the RPM display
uint8_t flapstatus = UP;                      // Flaps real position
uint8_t flaprequest = UP;                     // Flaps requested for flight condition f(Airspeed)
uint8_t gearstatus = UP;                      // Landing gear posistion
uint8_t gearrequest = UP;                     // Gear position requested for flight condition;
bool fuelcomputerflag = true;                 // flag used to print the fuel computer data
float oiltemp = 0.0;                          // Oil temperature (°C)
float oiltemp_Old = 0.0;                      // Oil Temperature FIR filter
float oilpress = 0.0;                         // Oil pressure (bar)
float oilpress_Old = 0.0;                     // Oil pressure FIR filter
float fuelpress = 0.0;                        // Fuel pressure (Bar)
float fuelpress_Old = 0.0;                    // Fuel pressure FIR filter
float fuelflow = 0;                           // Fuel  flow (lph)
float fuelflow_Old = 0;                       // Fuel  flow FIR filter
float lefttank = 0.0;                         // Calculated left tank level
float righttank = 0.0;                        // Calculated right tank level
float lefttank_Old = 0.0;                     // Calculated left tank level FIR filter
float righttank_Old = 0.0;                    // Calculated right tank level FIR filter
float egt[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };  // EGTs temperature (egt[i]=TC[i]+coldjtemp, i=0-7)
float egt_Old[8] = { 0, 0, 0, 0, 0, 0, 0, 0 }; // EGTs temperature FIR filter
float cht[2] = { 0, 0 };                    // CHTs temprature (aux1raw,aux2raw dependant from sensor type)
float cht_Old[2] = { 0, 0 };                // CHTs temprature FIR filter
float cat = 0.0;                              // Carburator Air Temperature
float cat_Old = 0.0;                          // Carburator Air Temperature
float volt = 0.0;                             // Battery voltage (volt = RDACvolt/5.73758)
float volt_Old = 0.0;                         // Battery voltage FIR filter
float ampbat = 0.0;                           // Battery current
float ampbus = 0.0;                           // Bus bar current
float ampbat_Old = 0.0;                       // Battery current FIR filter
float ampbus_Old = 0.0;                       // Bus bar current FIR filter
float Map = 0.0;                              // MAP Hg Inches
float Map_Old = 0.0;                          // MAP FIR filter
float mapold = 0.0;                           // update or not the MAP display
float vtemp = 0.0;                            // voltage temporary variable
float res = 0.0;                              // resistance temporary variable

alarm alm;

// Lookup table for PT1000 resistance-to-temperature conversion
float temperatureTable[] = { -40.0, -20.0, 0.0, 20.0, 40.0, 60.0, 80.0, 100.0, 120.0, 140.0, 160.0 };                // Corresponding temperature values
float resistanceTable[] = { 842.7, 921.6, 1000.0, 1077.9, 1155.4, 1232.4, 1309.0, 1385.1, 1460.7, 1535.8, 1610.4 };  // Add more values as needed

// Timers
elapsedMillis dispupdate1;  // display update timer
elapsedMillis dispupdate2;  // display update timer

// Engine timers struct

// ************************************************************** Variables CAN Bus ***************************************************************
CAN_message_t msgQNHaltitudesVario;
CAN_message_t msgSpeeds_Wind;
CAN_message_t msgIO_Enquiry;

// ************************************************ Variables utilisées pour les envois périodiques de données au CAN Bus **********************************
uint32_t periode200msStartTime;  // Variable utilisée pour les envois à 5 Hz
uint16_t periode200msTimeOut = 200;

// ************************************************************** Variables diverses **********************************************************************************************
int16_t backLightBrightness;

//------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                                                           SETUP
//------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void setup() {
  // *************************************************************************************** Initialisation de la carte microSD ****************************************************************************************************
  bool result = (SD.begin(BUILTIN_SDCARD));

  // *********************************************************************************** Initialisation du bus SPI1 et de l'écran  *******************************************************************************************
  SPI1.setMOSI(26);
  SPI1.setSCK(27);
  SPI1.setMISO(1);
  SPI1.setCS(0);
  SPI1.begin();
  tft.begin(Adafruit_800x480);
  tft.setRotation(3);
  tft.useLayers(true);
  tft.layerEffect(LAYER1);
  tft.writeTo(L1);
  tft.fillRect(0, 0, 480, 800, RA8875_BLACK);
  drawHourGlass(tft.width() / 2, tft.height() / 2);
  tft.writeTo(L2);
  tft.fillRect(0, 0, 480, 800, RA8875_BLACK);
  displayPhoto((char*)"MTBLANC2.out", 470, 262);  // affiche une photo sur la layer 2
  tft.layerEffect(LAYER2);
  tft.setCursor(41, 8);
  tft.setFont(Arial_24_Bold);
  tft.setTextColor(RA8875_WHITE);
  tft.println("EFIS AvionicsDuino v 4.00");
  tft.setCursor(40, 7);
  tft.setFont(Arial_24_Bold);
  tft.setTextColor(RA8875_BLUE);
  tft.println("EFIS AvionicsDuino v 4.00");
  tft.setFontDefault();
  tft.setFontScale(0);
  delay(199);
  tft.setTextColor(RA8875_GREEN);
  if (result) tft.println("Carte micro SD        : OK");
  else {
    tft.setTextColor(RA8875_RED);
    tft.println("Carte micro SD non disponible");
  }
  delay(199);
  tft.setTextColor(RA8875_GREEN);
  tft.println("LCD                   : OK");
  tft.println("PLEASE WAIT...");
  delay(7199);  // From 199 to 7199 due CAN bus error during RDAC startup
  
  // ************************************************************************ Initialisation du CAN bus  *****************************************************************************************************
  CAN_Module_EFIS.begin();
  CAN_Module_EFIS.setBaudRate(250000);
  CAN_Module_EFIS.setMaxMB(16);
  CAN_Module_EFIS.enableFIFO();
  CAN_Module_EFIS.enableFIFOInterrupt();
  CAN_Module_EFIS.onReceive(FIFO, canSniff);
  CAN_Module_EFIS.mailboxStatus();
  delay(199);

  // ********************************************************** Initialisation des messages CAN Bus ****************************************************************************
  msgQNHaltitudesVario.id = 46;
  msgQNHaltitudesVario.len = 8;
  msgSpeeds_Wind.id = 48;
  msgSpeeds_Wind.len = 8;
  msgIO_Enquiry.id = 1023; // 0x3FF Get IO input byte (Gear, flaps)
  msgIO_Enquiry.len = 8;
  for(uint8_t i=0;i<8;i++)
    msgIO_Enquiry.buf[i] = 0; // Arbitrary message does't care

  // ********************************************************************************* Initialisation des deux capteurs de pression  *******************************************************************************************
  if (AMS5915_050D.begin() < 0) {
    tft.setTextColor(RA8875_RED);
    tft.println("Erreur de communication avec le capteur AMS5915-50D");
  } else {
    tft.setTextColor(RA8875_GREEN);
    tft.println("Capteur AMS5915-50D   : OK");
  }
  delay(199);
  if (AMS5915_1500A.begin() < 0) {
    tft.setTextColor(RA8875_RED);
    tft.println("Erreur de communication avec le capteur AMS5915-1500A");
  } else {
    tft.setTextColor(RA8875_GREEN);
    tft.println("Capteur AMS5915-1500A : OK");
  }
  delay(199);

  // *********************************************************************** Initialisation du capteur de pression d'admission Adafruit MPRLS *************************************************
  if (!mpr.begin(0x18, &Wire2)) {
    tft.setTextColor(RA8875_RED);
    tft.println("Probleme capteur MAP");
    tft.println("Capteur MPRLS : OK");
  } else {
    tft.setTextColor(RA8875_GREEN);
    tft.println("Capteur MAP           : OK ");
  }
  delay(251);

  // ****************************************************************************** Initialisation des filtres des capteurs de pression *************************************************************************************
  AMS5915_1500A.readSensor('A');
  previousFilteredPressureValue = AMS5915_1500A.getPressure_Pa();
  altitudeVarioOld = int((1 - pow((((previousFilteredPressureValue + pressureCorrection * 10) / 100) / 1013.25), 0.190284)) * 145366.45);
  AMS5915_050D.readSensor('D');
  previousFilteredDiffPressureValue = AMS5915_050D.getPressure_Pa();
  tft.println("Filtres               : OK");
  delay(199);


  // ********************************************************************************* Initialisation des voies série  *******************************************************************************************
  Serial.begin(115200);
  tft.println("Serial                : OK");
  delay(199);

  // ********************************************************************************* Initialisations des timers *******************************************************************************************
  TimerClock.begin(setflagDisplayTime, 500ms);
  tft.println("Timers                : OK");
  delay(199);

  // *************************************************************************** Initialisation de l'encodeur rotatif ********************************************************************************************
  pinMode(Button_Pin, INPUT_PULLUP);
  attachInterrupt(Button_Pin, encoderButtonPress_ISR, RISING);
  encoder.setInitConfig();
  encoder.EncConfig.IndexTrigger = ENABLE;
  encoder.EncConfig.INDEXTriggerMode = RISING_EDGE;
  encoder.init();
  tft.println("Encodeur              : OK");
  delay(199);
  Button_Click = false;

  // *************************************************************************** Initialisation du capteur de température intérieure ********************************************************************************************
  if (!ds.search(addrDS18B20)) {
    tft.setTextColor(RA8875_RED);
    tft.println("Capteur de temperature intérieure non disponible");
    ds.reset_search();
    delay(199);
  }
  if (OneWire::crc8(addrDS18B20, 7) != addrDS18B20[7])
    tft.println("Le CRC de la ROM du capteur de température intérieure n'est pas valide !");
  else {
    tft.setTextColor(RA8875_GREEN);
    tft.println("CRC ROM DS18B20       : OK");
  }
  delay(199);

  // ********************************************************************************* Initialisation des variables stockées en EEPROM *************************************************************
  EEPROM.get(EFISeepromAdresse + 0, QNH);
  if ((QNH < 2750) || (QNH > 3150))  // Si cette condition est vraie (le QNH n'est pas dans l'intervalle 2750-3150, soit 931-1067 hPa), c'est a priori que l'EEPROM n'a pas été utilisée à cette adresse.
                                     // Il faut donc initialiser toutes les variables, puis les inscrire ensuite dans l'EEPROM
                                     // Si par la suite on règle volontairement le QNH à une valeur inférieure à 2750 (environ 930 hPa), toutes les valeurs stockées en EEPROM seront réinitialisées au prochain cycle.
  {
    QNH = 2992;
    magDev = 0;
    localTimeCorrection = 1;
    QFE = 2992;
    pressureCorrection = -20;
    backLightBrightness = 4;
    accGmax = 1.0;
    accGmin = 1.0;
    speedConversionFactor = 1.0;
    strcpy(strSpeedUnit, "Km/h");
    displayBaroSettingUnit = 0;
    displayTemperatureUnit = 0;

    EEPROM.put(EFISeepromAdresse + 0, QNH);                  // variable codée sur 2 octets
    EEPROM.put(EFISeepromAdresse + 2, magDev);               // variable codée sur 2 octets
    EEPROM.put(EFISeepromAdresse + 4, localTimeCorrection);  // variable codée sur 2 octets
    EEPROM.put(EFISeepromAdresse + 6, QFE);                  // variable codée sur 2 octets
    // Place ici pour une variable codée sur 2 octets, à l'adresse EFISeepromAdresse + 8 (by Carlo)
    EEPROM.put(EFISeepromAdresse + 8, flowDiv);                  // variable codée sur 2 octets
    EEPROM.put(EFISeepromAdresse + 10, pressureCorrection);      // variable codée sur 2 octets
    EEPROM.put(EFISeepromAdresse + 12, backLightBrightness);     // variable codée sur 2 octets
    EEPROM.put(EFISeepromAdresse + 14, accGmax);                 // variable codée sur 4 octets
    EEPROM.put(EFISeepromAdresse + 18, accGmin);                 // variable codée sur 4 octets
    EEPROM.put(EFISeepromAdresse + 22, speedConversionFactor);   // variable codée sur 4 octets
    EEPROM.put(EFISeepromAdresse + 30, strSpeedUnit);            // variable codée sur 5 octets
    EEPROM.put(EFISeepromAdresse + 35, displayBaroSettingUnit);  // variable codée sur 1 octet
    EEPROM.put(EFISeepromAdresse + 36, displayTemperatureUnit);  // variable codée sur 1 octet
    EEPROM.put(EFISeepromAdresse + 37, availablefuel);           // variable codée sur 1 octet
  } else {
    EEPROM.get(EFISeepromAdresse + 0, QNH);                  // variable codée sur 2 octets
    EEPROM.get(EFISeepromAdresse + 2, magDev);               // variable codée sur 2 octets
    EEPROM.get(EFISeepromAdresse + 4, localTimeCorrection);  // variable codée sur 2 octets
    EEPROM.get(EFISeepromAdresse + 6, QFE);                  // variable codée sur 2 octets
    // Place ici pour une variable codée sur 2 octets, à l'adresse EFISeepromAdresse + 8
    EEPROM.get(EFISeepromAdresse + 8, flowDiv);                  // variable codée sur 2 octets
    EEPROM.get(EFISeepromAdresse + 10, pressureCorrection);      // variable codée sur 2 octets
    EEPROM.get(EFISeepromAdresse + 12, backLightBrightness);     // variable codée sur 2 octets
    EEPROM.get(EFISeepromAdresse + 14, accGmax);                 // variable codée sur 4 octets
    EEPROM.get(EFISeepromAdresse + 18, accGmin);                 // variable codée sur 4 octets
    EEPROM.get(EFISeepromAdresse + 22, speedConversionFactor);   // variable codée sur 4 octets
    EEPROM.get(EFISeepromAdresse + 30, strSpeedUnit);            // variable codée sur 5 octets
    EEPROM.get(EFISeepromAdresse + 35, displayBaroSettingUnit);  // variable codée sur 1 octet
    EEPROM.get(EFISeepromAdresse + 36, displayTemperatureUnit);  // variable codée sur 1 octet
    EEPROM.get(EFISeepromAdresse + 37, availablefuel);           // variable codée sur 1 octet
  }
  tft.println("EEPROM                : OK");
  delay(199);

  // *************************************************** Démarrage de l'horloge à partir de l'heure conservée dans le SRTC par la pile *********************************************************************************************
  // Si la pile est morte, il faudra attendre le fix GPS pour avoir une heure.
  setSyncProvider(getTeensy3Time);
  setSyncInterval(900);
  tft.println("SRTC                  : OK");
  delay(199);

  // **************************************************************************************** Initialisation et tri des options des menus  **********************************************************************************************************************
  // Pour faciliter la gestion des options de menu (ajouts, suppressions, modifications), ces options sont affichées pendant le setup sur le terminal série, avant et après tri.
  uint8_t nb = numberOfMenuItems();
  Serial.println();
  Serial.println("EFIS AvionicsDuino version 4.00");
  Serial.print("Nombre de lignes utiles du tableau : ");
  Serial.println(nb);
  Serial.println("********Tableau avant tri et initialisation des index**********");
  Serial.println("Index----Label--------------Numero---Action---NbOptS--indexOfParentOption");
  for (byte n = 1; n <= nb + 1; n++) {
    Serial.printf("%2d\t", (n - 1));
    Serial.printf("%-8s\t", (menu[n - 1].menuItemLabel));
    Serial.printf("%7d\t\t", (menu[n - 1].menuItemNumber));
    Serial.print(menu[n - 1].menuItemAction);
    Serial.printf("\t%2d\t", (menu[n - 1].numberOfSisterOptions));
    Serial.printf("%2d", (menu[n - 1].indexOfParentOption));
    Serial.println();
  }
  bubbleSortTable();
  initTable();  // Initialisation des champs "numberOfSisterOptions" et indexOfParentOption"

  Serial.println();
  Serial.println("********Tableau après tri et initialisation des index**********");
  Serial.println("Index----Label--------------Numero---Action---NbOptS--indexOfParentOption");
  for (byte n = 1; n <= nb + 1; n++) {
    Serial.printf("%2d\t", (n - 1));
    Serial.printf("%-8s\t", (menu[n - 1].menuItemLabel));
    Serial.printf("%7d\t\t", (menu[n - 1].menuItemNumber));
    Serial.print(menu[n - 1].menuItemAction);
    Serial.printf("\t%2d\t", (menu[n - 1].numberOfSisterOptions));
    Serial.printf("%2d", (menu[n - 1].indexOfParentOption));
    Serial.println();
  }
  tft.println("Menus                 : OK");
  delay(199);

  // ********************************************************************************************** Démarrage de l'objet horizon ********************************************************************************************
  delay(1000);
  Horizon.begin();

  // ******************************************************* Dessin de l'horloge sur le premier plan (layer 1) et initialisations de quelques variables liées à cette horloge **************************************************
  tft.writeTo(L1);
  tft.setTextColor(RA8875_WHITE, RA8875_PINK);
  for (uint16_t i = 0; i < 360; i += 30) tft.drawLineAngle(Chx, Chy, i, Rh + 1, RA8875_WHITE);
  tft.fillCircle(Chx, Chy, Rh - 4, RA8875_PINK);
  hourHandLength = Rh * 2 / 3;
  minuteHandLength = Rh * 8 / 10;
  secondHandLength = Rh * 9 / 10;

  // ********************************************************************************* Draw the EMS frame *******************************************************************************************************************
  DrawEMSframe();
  ResetVars();
  // ********************************************************************************* On arrive presqu'à la fin de SETUP ***************************************************************************************************
  loopDurationStartTime = millis();
  baroVarioIntegrationStartTime = millis();

}  // ******************************************************************************************************************FIN DU SETUP***************************************************************************************************

//----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                                                                             LOOP
//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void loop() {
  tft.brightness(backLightBrightness * 4 - 1);
  tft.writeTo(L1);

  // Draw a red cross if AHRS messages are not received
  if (NoAHRSmsg > AHRS_MSGS_TIMEOUT && NoAHRSflag == false) {
    tft.fillQuad(1, 0, 479, 271, 478, 272, 0, 1, RA8875_RED);
    tft.fillQuad(0, 271, 478, 0, 479, 1, 1, 272, RA8875_RED);
    NoAHRSflag = true;
  }
  if (NoAHRSmsg == 0 && NoAHRSflag == true) {
    tft.fillQuad(1, 0, 479, 271, 478, 272, 0, 1, RA8875_PINK);
    tft.fillQuad(0, 271, 478, 0, 479, 1, 1, 272, RA8875_PINK);
    Horizon.reDraw();
    NoAHRSflag = false;
  }

  // Draw a red cross if RDAC messages are not received
  if (NoRDACmsg > RDAC_MSGS_TIMEOUT && NoRDACflag == false) {
    tft.fillQuad(2, 274, 479, 798, 478, 799, 0, 273, RA8875_RED);
    tft.fillQuad(0, 798, 478, 273, 479, 274, 1, 799, RA8875_RED);
    ResetVars();
    NoRDACflag = true;
  }
  if (NoRDACmsg == 0 && NoRDACflag == true) {
    tft.fillQuad(2, 274, 479, 798, 478, 799, 0, 273, RA8875_PINK);
    tft.fillQuad(0, 798, 478, 273, 479, 274, 1, 799, RA8875_PINK);
    DrawEMSframe();
    ResetVars();
    NoRDACflag = false;
  }
  if(NoRELAYmsg > RELAY_MSGS_TIMEOUT && NoRELAYflag == false)
  {
    tft.setFontScale(1);
    tft.setTextColor(RA8875_BLACK, RA8875_LIGHT_ORANGE);
    tft.setCursor(100, 355);
    tft.print("NO  DATA");
    tft.setFontScale(0);
    NoRELAYflag = true;
  }
  if(NoRELAYmsg == 0 && NoRELAYflag == true)
  {
    tft.setFontScale(1);
    tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
    tft.setCursor(100, 355);
    tft.print("        ");
    tft.setFontScale(0);
    NoRELAYflag = false;
  }

  NoAHRSmsg++;
  NoRDACmsg++;
  NoFUELmsg++;
  NoRELAYmsg++;

  // ********************************************** Réglage de l'heure sur celle du GNSS de l'AHRS, une fois qu'il diffuse une heure exacte **********************************************************
  if (!flagTimeSet) {
    if ((timeAccuracyGNSS < 1000) && (yearGNSS > 0))  // Cold Start : le GNSS commence à diffuser la date et l'heure à partir de 8 satellites en vue,
                                                      // mais il faut attendre environ 14 ou 15 satellites pour que la précision de l'heure soit < 1000 ns, ce qui peut être long...
                                                      // Hot start : la précision peut être d'emblée < 1000 ns, mais la diffusion de la date et et de l'heure peut prendre ensuite plusieurs longues secondes.
                                                      // Il faut donc attendre impérativement que les deux conditions (timeAccuracyGNSS < 1000) && (yearGNSS>0)soient réunies pour régler l'heure.
    {
      setTime(hourGNSS, minuteGNSS, secondGNSS, dayGNSS, monthGNSS, yearGNSS);  // L'heure de la bibliothèque Time est désormais l'heure UTC
      Teensy3Clock.set(now());                                                  // Et cette même heure est desormais aussi celle du RTC/SRTC, sauvegardée par la pile.
      flagTimeSet = true;
    }
  }

  // ****************************************************************************************** Lecture des capteurs de pression ****************************************************************************************************
  AMS5915_050D.readSensor('D');
  AMS5915_1500A.readSensor('A');
  pressure = AMS5915_1500A.getPressure_Pa();
  nonFilteredPressure = pressure;
  pressure = iirFilter(previousFilteredPressureValue, pressure, 0.005);
  previousFilteredPressureValue = pressure;
  differentialPressure = AMS5915_050D.getPressure_Pa();
  nonFilteredDifferentialPressure = differentialPressure;
  differentialPressure = iirFilter(previousFilteredDiffPressureValue, differentialPressure, 0.005);
  previousFilteredDiffPressureValue = differentialPressure;

  // ****************************************************************************************** Lecture du capteur de température intérieure *********************************************************************************************
  if (flagDS18B20AcquisitionDelayElapsed == true) {
    ds.reset();
    ds.select(addrDS18B20);
    ds.write(0x44);
    DS18B20AcquisitionStartTime = millis();
    flagDS18B20AcquisitionDelayElapsed = false;
  }
  if ((millis() - DS18B20AcquisitionStartTime) > 1000) {
    ds.reset();
    ds.select(addrDS18B20);
    ds.write(0xBE);
    for (byte i = 0; i < 9; i++) dataDS18B20[i] = ds.read();
    temperatureDS18B20 = (float)((dataDS18B20[1] << 8) | dataDS18B20[0]) / 16.0;
    flagDS18B20AcquisitionDelayElapsed = true;
  }

  //******************************************************************************* Acquisition de la pression d'admission *********************************************************************************
  //manifoldPressure = mpr.readPressure()*0.02953; // MAP acquise en hPa, puis convertie en Inch Hg
  manifoldPressure = mpr.readPressure();  // MAP acquise en deci milliBar
  nonFilteredManifoldPressure = manifoldPressure;
  manifoldPressure = iirFilter(previousFilteredManifoldPressureValue, manifoldPressure, 0.03);
  previousFilteredManifoldPressureValue = manifoldPressure;
  Map = manifoldPressure * 0.02953;  // Convet from mBar to HgInches
  // ************************************************************************************ Calculs des affichages textuels *********************************************************************************************************

  // --------- calcul des altitudes --------- (Les pressions sont exprimées en Pascals, et filtrées dès l'étape de lecture des capteurs, mais les QNH/QFE sont exprimés en centièmes de pouces de mercure)
  qnhAltitude = (1 - pow((((pressure + pressureCorrection * 10) / 100) / (QNH * 0.338646)), 0.190284)) * 145366.45 * altitudeCorrection;  // 0.01 inHg = 0.338646 hPa
  qnhAltitude = int((qnhAltitude + 5) / 10.0) * 10;
  qfeAltitude = (1 - pow((((pressure + pressureCorrection * 10) / 100) / (QFE * 0.338646)), 0.190284)) * 145366.45 * altitudeCorrection;
  qfeAltitude = int((qfeAltitude + 5) / 10.0) * 10;
  pressureAltitude = (1 - pow((((pressure + pressureCorrection * 10) / 100) / 1013.25), 0.190284)) * 145366.45 * altitudeCorrection;
  altitudeVario = pressureAltitude;
  pressureAltitude = int((pressureAltitude + 5) / 10.0) * 10;
  iat = temperatureDS18B20;
  if (iat > 1000) iat = 20.0;
  float pressionSatH20 = 6.1078 * pow(10, ((7.5 * oat) / (237.3 + oat))) * 100;
  float pressionPartielleH2O = pressionSatH20 * relativeHumidity / 100;
  float densiteAir = ((pressure + pressureCorrection * 10) / (287.05 * (oat + 273.15)) * (1 - (0.378 * pressionPartielleH2O / (pressure + pressureCorrection * 10))));
  densityAltitude = (44.3308 - (42.2665 * (pow(densiteAir, 0.234969)))) * 1000 * 3.28084;
  densityAltitude = int((densityAltitude + 5) / 10.0) * 10;
  dewPoint = 243.12 * (log(relativeHumidity / 100) + 17.62 * oat / (243.12 + oat)) / (17.62 - (log(relativeHumidity / 100) + 17.62 * oat / (243.12 + oat)));

  // ----------- Calcul du taux de montée/descente à partir du capteur de pression statique -------------------
  if (millis() >= baroVarioIntegrationStartTime + baroVarioIntegrationTimeOut) {
    baroVario = (altitudeVario - altitudeVarioOld) * 60 * (1000 / baroVarioIntegrationTimeOut);
    baroVarioIntegrationStartTime = millis();
    altitudeVarioOld = altitudeVario;
    baroVario = iirFilter(previousFilteredBaroVarioValue, baroVario, 0.05);
    previousFilteredBaroVarioValue = baroVario;
    if (baroVario >= 0) baroVario = (int(baroVario + 5) / 10) * 10;
    else baroVario = (int(baroVario - 5) / 10) * 10;
  }

  // ----------------------- calcul des vitesses de l'avion, résultat dans l'unité sélectionnée par l'utilisateur --------------------------------------------
  if (differentialPressure > 4.5) {
    indicatedAirSpeed = int((sqrtf(21.159183 * differentialPressure) / speedConversionFactor) + 0.5);                         // Cette valeur est fortement filtrée, car la variable differentialPressure l'est également
    trueAirSpeed = int((86.257392 * sqrtf(differentialPressure * (273.15 + oat) / pressure) / speedConversionFactor) + 0.5);  // Cette valeur est fortement filtrée, car la variable differentialPressure l'est également

    nonFilteredIndicatedAirSpeed = int(sqrtf(21.159183 * nonFilteredDifferentialPressure) + 0.5);                         // Valeur non filtrée en km/h pour envoi au CAN bus, et donc au FDR
    nonFilteredTrueAirSpeed = int(86.257392 * sqrtf(nonFilteredDifferentialPressure * (273.15 + oat) / pressure) + 0.5);  // Valeur non filtrée en km/h pour envoi au CAN bus, et donc au FDR
  } else {
    indicatedAirSpeed = 0;
    trueAirSpeed = 0;
  }

  // ----------------------- calcul direction et vitesse du vent (nécessite un magnétomètre PARFAITEMENT calibré et aligné !!!) --------------------------------------------
  float GSE = sin(PI / 180 * trk) * groundSpeedGNSS;  // groundSpeedGNSS est exprimée en Km/h
  float GSN = cos(PI / 180 * trk) * groundSpeedGNSS;
  float TASE = sin(PI / 180 * (magneticHeading + magDev)) * trueAirSpeed;
  float TASN = cos(PI / 180 * (magneticHeading + magDev)) * trueAirSpeed;
  float deltaE = TASE - GSE;
  float deltaN = TASN - GSN;
  windSpeed = sqrt(pow(deltaE, 2) + pow(deltaN, 2));  // résultat en km/h
  windDirection = atan2(deltaE, deltaN) * 180 / PI;
  if (windDirection < 0) windDirection = 360 + windDirection;

  // **************************************************************************************** Mise à l'heure de la pendule et des chronos *********************************************************************************************
  if (flagDisplayTime) {
    flagDisplayTime = false;
    displayTime();
    time_t i;
    if (flag_chrono1) {
      i = now() - t_chrono1;
      tft.setCursor(80, 222);
      tft.printf("%02u:%02u:%02u", hour(i), minute(i), second(i));
    }
    if (flag_chrono2) {
      i = now() - t_chrono2;
      tft.setCursor(180, 222);
      tft.setTextColor(RA8875_GREEN, RA8875_BLACK);
      tft.printf("%02u:%02u", minute(i), second(i));
      tft.setTextColor(RA8875_WHITE, RA8875_PINK);
    }
    if (flagTimer) {
      i = t_timer - now();
      if (i >= 0) {
        tft.setTextColor(RA8875_WHITE, RA8875_RED);
        tft.setCursor(260, 222);
        tft.printf("%02u:%02u", minute(i), second(i));
        tft.setTextColor(RA8875_WHITE, RA8875_PINK);
      } else {
        flagTimer = false;
        tft.setCursor(260, 222);
        tft.print("     ");
      }
    }
  }

  // ************************************************************************************ Mise à jour des affichages textuels et du vario ****************************************************************************************
  displayTexts();
  displayVario();

  //  ***************************************************************************** Traitement d'un clic sur le bouton de l'encodeur *******************************************************************************************

  if (Button_Click == true) {
    if (Menu_open == false) {
      Menu_open = true;
      openMenu(0);
    } else {
      if (SetVal == false) {
        switch (menu[indexCurrentMenuOption].menuItemAction) {
          case 'Q':
            Menu_open = false;
            closeMenu();
            break;
          case 'S':
            openMenu(menu[indexCurrentMenuOption].indexOfParentOption);
            break;
          case 'B':
            openMenu(menu[indexCurrentMenuOption].indexOfParentOption);
            ;
            break;
          case 'P':
            SetVal = true;
            encoder.write(0);
            EncoderPositionValueOld = EncoderPositionValue;
            EncoderPositionValueOldSetVal = 0;
            Button_Click = false;
            switch (menu[indexCurrentMenuOption].menuItemNumber) {
              case 101:
                ptrGen_int16_t = &QNH;
                OffsetEFISeepromAdresse = 0;
                break;
              case 102:
                ptrGen_int16_t = &QFE;
                OffsetEFISeepromAdresse = 6;
                break;
              case 112:
                ptrGen_int16_t = &magDev;
                if(*ptrGen_int16_t > 30)
                  *ptrGen_int16_t = 30;
                if(*ptrGen_int16_t < -30)
                  *ptrGen_int16_t = -30;

                OffsetEFISeepromAdresse = 2;
                break;
              case 113:
                ptrGen_int16_t = &pressureCorrection;
                OffsetEFISeepromAdresse = 10;
                break;
              case 114:
                ptrGen_int16_t = &backLightBrightness;
                OffsetEFISeepromAdresse = 12;
                break;
              case 115:
                ptrGen_int16_t = &localTimeCorrection;
                OffsetEFISeepromAdresse = 4;
                break;
              case 123:
                ptrGen_int16_t = &timerDuration;
                OffsetEFISeepromAdresse = 1000;
                flagTimer = false;
                tft.setCursor(320, 222);
                tft.print("     ");
                break;
              case 131:
                ptrGen_int16_t = &flowDiv;
                OffsetEFISeepromAdresse = 8;
                break;
              case 132:
                ptrGen_int16_t = &availablefuel;
                if(*ptrGen_int16_t > 72)
                  *ptrGen_int16_t = 72;
                OffsetEFISeepromAdresse = 37;
                break;
              default: break;
            }
            tft.fillRect((((menu[indexCurrentMenuOption].menuItemNumber) % 10) * boxWidth), (menuYCoordinate - boxHeight + 3), boxWidth, boxHeight - 3, RA8875_RED);
            tft.setTextColor(RA8875_WHITE, RA8875_RED);
            tft.setFont(Arial_10);
            tft.setCursor(((menu[indexCurrentMenuOption].menuItemNumber) % 10) * boxWidth + (boxWidth / 4), menuYCoordinate + 4 - boxHeight);
            if (ptrGen_int16_t == &QNH || ptrGen_int16_t == &QFE) {
              if (displayBaroSettingUnit == 1)  // alors on va afficher le QNH/QFE en pouces de mercure
              {
                tft.printf("%02u.%02u", (*ptrGen_int16_t / 100), (*ptrGen_int16_t % 100));
              } else  // sinon on va afficher le QNH/QFE en hPa
              {
                float i = *ptrGen_int16_t * 0.338646;
                uint16_t j = int(i + 0.5);
                tft.printf("% 5d", j);
              }
            } else tft.print(*ptrGen_int16_t);
            tft.setFont(Arial_11);
            break;  // break du case 'P'

          case 'A':
            switch (menu[indexCurrentMenuOption].menuItemNumber) {
              case 103:
                accGmax = 1.0;
                accGmin = 1.0;
                EEPROM.put(EFISeepromAdresse + 14, accGmax);
                EEPROM.put(EFISeepromAdresse + 18, accGmin);
                openMenu(indexFirstSisterOption(indexCurrentMenuOption));
                break;
              case 121:
                if (flag_chrono1 == false) {
                  t_chrono1 = now();
                  flag_chrono1 = true;
                } else {
                  tft.setCursor(80, 222);
                  tft.print("        ");
                  flag_chrono1 = false;
                }
                break;

              case 122:
                if (flag_chrono2 == false) {
                  t_chrono2 = now();  // Si le chrono2 ne tourne pas, on le démarre
                  flag_chrono2 = true;
                } else {
                  tft.setCursor(220, 222);  // S'il tourne, on l'efface et on l'arrête
                  tft.print("        ");
                  flag_chrono2 = false;
                }
                break;
              case 1111:
                speedConversionFactor = 1.0;
                strcpy(strSpeedUnit, "Km/h");
                EEPROM.put(EFISeepromAdresse + 22, speedConversionFactor);
                EEPROM.put(EFISeepromAdresse + 30, strSpeedUnit);
                openMenu(indexFirstSisterOption(indexCurrentMenuOption));
                break;
              case 1112:
                speedConversionFactor = 1.852;
                strcpy(strSpeedUnit, "Kts ");
                EEPROM.put(EFISeepromAdresse + 22, speedConversionFactor);
                EEPROM.put(EFISeepromAdresse + 30, strSpeedUnit);
                openMenu(indexFirstSisterOption(indexCurrentMenuOption));
                break;
              case 1113:
                speedConversionFactor = 1.609;
                strcpy(strSpeedUnit, "mph ");
                EEPROM.put(EFISeepromAdresse + 22, speedConversionFactor);
                EEPROM.put(EFISeepromAdresse + 30, strSpeedUnit);
                openMenu(indexFirstSisterOption(indexCurrentMenuOption));
                break;
              case 1114:
                displayBaroSettingUnit = 0;
                EEPROM.put(EFISeepromAdresse + 35, displayBaroSettingUnit);
                openMenu(indexFirstSisterOption(indexCurrentMenuOption));
                break;
              case 1115:
                displayBaroSettingUnit = 1;
                EEPROM.put(EFISeepromAdresse + 35, displayBaroSettingUnit);
                openMenu(indexFirstSisterOption(indexCurrentMenuOption));
                break;
              case 1116:
                displayTemperatureUnit = 0;
                EEPROM.put(EFISeepromAdresse + 36, displayBaroSettingUnit);
                openMenu(indexFirstSisterOption(indexCurrentMenuOption));
                break;
              case 1117:
                displayTemperatureUnit = 1;
                EEPROM.put(EFISeepromAdresse + 36, displayBaroSettingUnit);
                openMenu(indexFirstSisterOption(indexCurrentMenuOption));
                break;
            }
            break;
        }
      }

      else  // On est dans le cas où le clic vient valider la modification d'une valeur numérique
      {
        if (OffsetEFISeepromAdresse != 1000) {
          EEPROM.put(EFISeepromAdresse + OffsetEFISeepromAdresse, *ptrGen_int16_t);
        }
        if (timerDuration <= 0) {
          flagTimer = false;
          tft.setCursor(320, 222);
          tft.print("     ");
        } else if (!flagTimer) {
          t_timer = now() + timerDuration * 60;
          flagTimer = true;
        }
        SetVal = false;
        ptrGen_int16_t = NULL;
        openMenu(indexFirstSisterOption(indexCurrentMenuOption));  // et on revient à la première option du menu en cours
        //closeMenu(); // Ou, parfois plus ergonomique, on ferme le menu. En fait, un bouton séparé pour la fermeture du menu serait souhaitable.
      }
    }
    Button_Click = false;
  }

  // ************************************************************************* Polling de la position de l'encodeur à chaque passage dans Loop ************************************************************************************
  if (Menu_open == true) {
    if (SetVal == false) {
      EncoderPositionValue = encoder.read() >> 2;
      tft.setTextColor(RA8875_WHITE, RA8875_PINK);
      if (EncoderPositionValue != EncoderPositionValueOld) {
        uint16_t minimum = menu[indexCurrentMenuOption].menuItemNumber;
        minimum = minimum % 10;
        minimum = indexCurrentMenuOption - minimum;
        uint8_t maximum = minimum + menu[minimum].numberOfSisterOptions - 1;
        if ((EncoderPositionValue > maximum - minimum) || (EncoderPositionValue < 0)) {
          encoder.write(0);
          EncoderPositionValue = encoder.read() >> 2;
        }
        indexCurrentMenuOption = EncoderPositionValue + minimum;
        tft.drawRect((EncoderPositionValueOld * boxWidth), menuYCoordinate, boxWidth, boxHeight, BckgrndMenus);
        tft.drawRect((EncoderPositionValue * boxWidth), menuYCoordinate, boxWidth, boxHeight, RA8875_WHITE);
      }
      EncoderPositionValueOld = EncoderPositionValue;
    } else {
      EncoderPositionValue = encoder.read() >> 2;
      if (EncoderPositionValue != EncoderPositionValueOldSetVal) {
        int8_t variation = encoder.getHoldDifference();
        if (displayBaroSettingUnit == 0 && (ptrGen_int16_t == &QNH || ptrGen_int16_t == &QFE))  // si on incrémente QNH/QFE en hPa, alors il faut incrémenter x3
        {
          *ptrGen_int16_t = *ptrGen_int16_t + (3 * variation);
        } else *ptrGen_int16_t = *ptrGen_int16_t + variation;
        validateValues();
        tft.fillRect((((menu[indexCurrentMenuOption].menuItemNumber) % 10) * boxWidth), (menuYCoordinate - boxHeight + 3), boxWidth, boxHeight - 3, RA8875_RED);
        tft.setTextColor(RA8875_WHITE, RA8875_RED);
        tft.setFont(Arial_10);
        tft.setCursor(((menu[indexCurrentMenuOption].menuItemNumber) % 10) * boxWidth + (boxWidth / 4), menuYCoordinate + 4 - boxHeight);
        if (ptrGen_int16_t == &QNH || ptrGen_int16_t == &QFE) {
          if (displayBaroSettingUnit == 1)  // alors on va afficher le QNH/QFE en pouces de mercure
          {
            tft.printf("%02u.%02u", (*ptrGen_int16_t / 100), (*ptrGen_int16_t % 100));
          } else  // sinon on va afficher le QNH/QFE en hPa
          {
            float i = *ptrGen_int16_t * 0.338646;
            uint16_t j = int(i + 0.5);
            tft.printf("% 5d", j);
          }
        } else tft.print(*ptrGen_int16_t);
        tft.setFont(Arial_11);
      }
      EncoderPositionValueOldSetVal = EncoderPositionValue;
    }
  }

  // ************************************************************************************ Mise à jour de l'horizon artificiel****************************************************************************************
  ballOffset = (accY * -1 * ballRadius) / 0.27;
  if (ballOffset > ballRadius * 8) ballOffset = ballRadius * 8;
  if (ballOffset < -ballRadius * 8) ballOffset = -ballRadius * 8;
  Horizon.draw(roll, pitch, 0, ballOffset, Menu_open);  // roll en radians, pitch en degrés, la valeur 0 est en réserve pour l'avenir, inutile pour le moment.

  // ************************************************************************* Draw all Engine (EMS) parameters ***************************************************************************************************
  if (rpm < 500 && abs(manifoldPressure - pressure / 100) > MAXPRESDELTA && millis() > 25000 && millis() < 30000) {
    tft.setTextColor(RA8875_RED, RA8875_YELLOW);
    tft.setCursor(260, 735);
    tft.print("    WARNING MAP VS QNH    ");
    tft.setCursor(260, 750);
    tft.printf("MAP: % 2.1f VS QNH: % 2.1f", manifoldPressure, pressure / 100);
    tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  }

  DrawRpmMap();

  if (dispupdate2 > 5000)  // update display every 5 seconds
  {
    dispupdate2 = 0;
    DrawCAT();
    DrawEgt();
    DrawCht();
    DrawOilT();
    DrawTanks();
    DrawFuelComp();
  }

  if (dispupdate1 > 1000)  // update display every 1 second
  {
    dispupdate1 = 0;
    CAN_Module_EFIS.write(msgIO_Enquiry);
    DrawFlapGear();
    DrawOilP();
    DrawFuelP();
    DrawElectric();
    DrawMessage();
  }
  
  ShowAlarms(); // Check if some alarm exists

  tft.setTextColor(RA8875_WHITE, RA8875_PINK);
  tft.setFontScale(0);

  // ************************************************************************* Mesure de la durée de la boucle principale Loop ************************************************************************************
  if (loopCounter >= numLoopIteration) {
    numLoopIterationDuration = millis() - loopDurationStartTime;
    loopDuration = float(numLoopIterationDuration / numLoopIteration);
    //tft.setCursor(450,257);
    //tft.print(loopDuration,1);
    loopDurationStartTime = millis();
    loopCounter = 0;
  } else {
    loopCounter++;
  }
}

// ***************************************************************************************************** Fin de la boucle infinie Loop ***********************************************************************************************

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                                                   ISR de récéption du CAN Bus
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void canSniff(const CAN_message_t& msg) {
  switch (msg.id) {
    case 42:
      oat = *(float*)(msg.buf + 0);
      break;

    case 44:
      relativeHumidity = *(float*)(msg.buf + 0);
      break;

    case 28:
      magneticHeading = *(float*)(msg.buf + 0);
      if ((magneticHeading < 0) && (previousFilteredMagneticHeadingValue < 90)) {
        previousFilteredMagneticHeadingValue = 360;
      }
      if ((magneticHeading > 0) && (previousFilteredMagneticHeadingValue > 270)) {
        previousFilteredMagneticHeadingValue = 0;
      }
      if (magneticHeading < 0) magneticHeading += 360;
      magneticHeading = iirFilter(previousFilteredMagneticHeadingValue, magneticHeading, 0.07);
      previousFilteredMagneticHeadingValue = magneticHeading;
      intMagneticHeading = uint16_t(magneticHeading + 0.5);
      break;

    case 20:
      roll = *(float*)(msg.buf + 0);  // L'AHRS transmet au CAN Bus un angle de roll non filtré, en radians. Il faut filtrer pour avoir un horizon stable.
      if (roll < 2.8 && roll > -2.8)  // Mais au voisinage du passage de 180° à -180° ( entre 2.8 radians et -2.8 radians soit entre 160° et -160°) il ne faut pas filtrer, pour éviter une bascule brutale à 180° de l'horizon.
      {
        roll = iirFilter(previousFilteredRollValue, roll, 0.1);
      }
      previousFilteredRollValue = roll;

      pitch = (*(float*)(msg.buf + 4));  // L'AHRS transmet au CAN Bus un angle de pitch non filtré, en degrés. Il faut filtrer pour avoir un horizon stable.
      pitch = iirFilter(previousFilteredPitchValue, pitch, 0.1);
      previousFilteredPitchValue = pitch;
      NoAHRSmsg = 0;
      break;

    case 22:
      accY = *(float*)(msg.buf + 0);  // L'AHRS transmet au CAN Bus une valeur accY non filtré. Il faut filtrer pour avoir une bille stable.
      accY = iirFilter(previousFilteredAccYValue, accY, 0.02);
      previousFilteredAccYValue = accY;

      accG = *(float*)(msg.buf + 4);
      accG = -1 * accG / 9.80665;
      if ((accG > accGmax) && (differentialPressure >= 380))  // on ne tient pas compte des G transitoires liés à d'éventuelles secousses durant un roulage lent
      {
        accGmax = accG;
        EEPROM.put(EFISeepromAdresse + 14, accGmax);
      }
      if ((accG < accGmin) && (differentialPressure >= 380)) {
        accGmin = accG;
        EEPROM.put(EFISeepromAdresse + 18, accGmin);
      }
      accG = iirFilter(previousFilteredAccGValue, accG, 0.1);  // Filtrage modéré
      previousFilteredAccGValue = accG;
      NoAHRSmsg = 0;
      break;

    case 24:
      Vz = *(float*)(msg.buf + 0);                        // L'AHRS transmet au CAN Bus une valeur de Vz float non filtrée, en m/s.
      Vz = iirFilter(previousFilteredVzValue, Vz, 0.03);  // Cette valeur de Vz, en m/s est filtrée pour la stabilité de l'affichage,
      previousFilteredVzValue = Vz;
      Vz = Vz * 196.8504;  // puis convertie en ft/min,
      if (Vz >= 0) {
        roundedIntVz = (int(Vz + 5) / 10) * 10;  // et enfin arrondie à la dizaine la plus proche
      } else {
        roundedIntVz = (int(Vz - 5) / 10) * 10;
      }

      trk = *(float*)(msg.buf + 4);                      // L'AHRS transmet au CAN Bus une valeur de trk float non filtrée, en degrés, de -180° à +180°.
      if ((trk < 0) && (previousFilteredTrkValue < 90))  // Précaution pour le filtrage aux alentours de 360°
      {
        previousFilteredTrkValue = 360;
      }
      if ((trk > 0) && (previousFilteredTrkValue > 270))  // Précaution pour le filtrage aux alentours de 0°
      {
        previousFilteredTrkValue = 0;
      }
      if (trk < 0) trk = 360 + trk;
      trk = iirFilter(previousFilteredTrkValue, trk, 0.1);  // Filtre IIR
      previousFilteredTrkValue = trk;                       // On a maintenant une valeur de trk float, filtrée, en degrés, de 0 à 360°.
      NoAHRSmsg = 0;
      break;

    case 34:
      yearGNSS = *(uint16_t*)(msg.buf + 0);
      monthGNSS = *(uint8_t*)(msg.buf + 2);
      dayGNSS = *(uint8_t*)(msg.buf + 3);
      hourGNSS = *(uint8_t*)(msg.buf + 4);
      minuteGNSS = *(uint8_t*)(msg.buf + 5);
      secondGNSS = *(uint8_t*)(msg.buf + 6);
      satInView = *(uint8_t*)(msg.buf + 7);
      NoAHRSmsg = 0;
      break;

    case 38:
      nanoSecGNSS = *(int32_t*)(msg.buf + 0);
      timeAccuracyGNSS = *(uint32_t*)(msg.buf + 4);
      NoAHRSmsg = 0;
      break;

    case 30:
      altitudeGNSS = *(float*)(msg.buf + 0);                          // altitude MSL passée en pieds par l'AHRS au can bus, non filtrée.
      altitudeGNSS = iirFilter(altitudeGNSS_Old, altitudeGNSS, 0.1);  // Filtrage IIR
      altitudeGNSS_Old = altitudeGNSS;

      groundSpeedGNSS = *(float*)(msg.buf + 4);                                // vitesse sol passée en km/h par l'AHRS au can bus, non filtrée
      groundSpeedGNSS = iirFilter(groundSpeedGNSS_Old, groundSpeedGNSS, 0.1);  // filtrage IIR
      groundSpeedGNSS_Old = groundSpeedGNSS;
      NoAHRSmsg = 0;
      break;

    case 0x3FF:  // I/O module inputs 0-3 (response to status request: see line 1141 "CAN_Module_EFIS.write(msgIO_Enquiry)")
      gearstatus = (*(uint8_t*)(msg.buf + 0))>>GEARFLAG & 0x01;
      flapstatus = (*(uint8_t*)(msg.buf + 0))>>FLAPFLAG & 0x01;
      NoRELAYmsg = 0;
      break;

    default: break;
  }
  if ((msg.id) >> 4 == RDAC_CAN_ID) {
    switch ((msg.id) & 0x0F)
    {
      case 1:
        fuelflowcount = *(uint16_t*)(msg.buf + 0);  // Pulses within last 4 seconds ID 1, 0-1
        fuelflowratio = *(uint16_t*)(msg.buf + 2);  // Pulses duty cycle x10 ID 1, 2-3 (400 = 60.0/40.0% ratio) No pulse detected = 65535
        fuelflow = (float)fuelflowcount * 900.0 / flowDiv;
        fuelpulse += fuelflowcount;
        fuelburnt = fuelpulse / flowDiv;
        fuelcomputerflag = true;
        break;
      case 2:
        for (uint8_t i = 0; i < 4; i++)
        {
          tc[i] = *(uint16_t*)(msg.buf + (i * 2));
          if (coldjtemp > -100)
          {
            egt[i] = (float)(tc[i] + coldjtemp);
            egt[i] = iirFilter(egt_Old[i], egt[i], 0.1);  // FIR filter 10%)
            egt_Old[i] = egt[i];
          }
        }
        break;
      case 3:
        for (uint8_t i = 4; i < 8; i++)
        {
          tc[i] = *(uint16_t*)(msg.buf + (i * 2));
          if (coldjtemp > -100)
          {
            egt[i] = (float)(tc[i] + coldjtemp);
            egt[i] = iirFilter(egt_Old[i], egt[i], 0.1);  // FIR filter 10%)
            egt_Old[i] = egt[i];
          }
        }
        break;
        //      case 4: non implemented for RDAC XG

      case 5:
        oiltraw = *(uint16_t*)(msg.buf + 0);
        oilpraw = *(uint16_t*)(msg.buf + 2);
        aux1raw = *(uint16_t*)(msg.buf + 4);
        aux2raw = *(uint16_t*)(msg.buf + 6);

        // OIL temperature
        vtemp = ((float)oiltraw / (float)ADCMAX) * VREF;
        res = (vtemp * 1500.0) / (VREF - vtemp);
        oiltemp = VDOr2t(res);
        //oiltemp = interpolate(res, resistanceTable, temperatureTable, sizeof(resistanceTable) / sizeof(resistanceTable[0]));
        oiltemp = iirFilter(oiltemp_Old, oiltemp, 0.1);  // FIR filter 10%
        oiltemp_Old = oiltemp;

        // Oil Pressure
        oilpress = ((float)(oilpraw -324)/(float)(1630 - 324)) * 10.0;
        oilpress = iirFilter(oilpress_Old, oilpress, 0.1);  // FIR filter 10%
        oilpress_Old = oilpress;

        // CHT1 temperature
        vtemp = ((float)aux1raw / ADCMAX) * VREF;
        res = (vtemp * 1500) / (VREF - vtemp);
        cht[0] = (int16_t)round(VDOr2t(res));
        cht[0] = iirFilter(cht_Old[0], cht[0], 0.1);  // FIR filter 10%)
        cht_Old[0] = cht[0];

        // CHT2 temperature
        vtemp = ((float)aux2raw / ADCMAX) * VREF;
        res = (vtemp * 1500) / (VREF - vtemp);
        cht[1] = (int16_t)round(VDOr2t(res));
        cht[1] = iirFilter(cht_Old[1], cht[1], 0.1);  // FIR filter 10%)
        cht_Old[1] = cht[1];
        break;

      case 6:
        fuelpraw = *(uint16_t*)(msg.buf + 0);
        coolraw = *(uint16_t*)(msg.buf + 2);
        fuellev1raw = *(uint16_t*)(msg.buf + 4);
        fuellev2raw = *(uint16_t*)(msg.buf + 6);

        // Fuel Pressure
        //fuelpress = ((float)(fuelpraw -324)/(float)(1630 - 324)) * 10.0; //  to be updated
        fuelpress = ((float)fuelpraw * VREF /  (float)ADCMAX) - FPOFFSET;
    		fuelpress=((fuelpress*FPGAIN)-0.4)*1.5625;


//        fuelpress = iirFilter(fuelpress_Old, fuelpress, 0.1);  // FIR filter 10%)
//        fuelpress_Old = fuelpress;

        // The coolant sensor has not been used, bu it can be activated if needed
        // so, the coolraw is a spare
        // The FL1 input will be used to read a second current sensor
        ampbus = (((float)fuellev1raw * VREF / (float)ADCMAX) - AMPOFFS1) / AMPVPA1;
        ampbus = iirFilter(ampbus_Old, ampbus, 0.1);  // filtrage IIR
        ampbus_Old = ampbus;

        // The FL2 input will be used to read the CAT by a LM335
        cat = ((float)fuellev2raw * VREF / (float)ADCMAX) * 100 - 273.15;
        cat = iirFilter(cat_Old, cat, 0.1);  // filtrage IIR
        cat_Old = cat;
        break;

      case 7:
        coldjtemp = *(uint16_t*)(msg.buf + 0);
        RDACvolt = *(uint16_t*)(msg.buf + 2);
        volt = round(((float)RDACvolt) / (float)5.73758) / 10.0;
        // Under 14V the RDAC's power supply looses about 0.9833% of voltage
        if (volt < 14.00)
          volt = volt / 0.98333;
        volt = iirFilter(volt_Old, volt, 0.1);  // filtrage IIR
        volt_Old = volt;
        
        break;

      case 8:
        rpm = *(uint16_t*)(msg.buf + 0);
        //rpm2 = *(uint16_t*)(msg.buf+2); // Not for RDAC XG
        //map = *(uint16_t*)(msg.buf+4); // Not for RDAC XG
        curraw = *(uint16_t*)(msg.buf + 6);
        ampbat = ((((float)curraw) * VREF / (float)ADCMAX) - AMPOFFS2) / AMPVPA2;
        ampbat = iirFilter(ampbat_Old, ampbat, 0.1);  // filtrage IIR
        ampbat_Old = ampbat;
        break;

      default: break;
    }
    NoRDACmsg = 0;
  }

  if ((msg.id) >> 4 == FUEL_CAN_ID) {
    switch ((msg.id) & 0x0F) {
      case 0:
        lefttankraw = (*(uint16_t*)(msg.buf + 0));
        righttankraw = (*(uint16_t*)(msg.buf + 2));
        auxtankraw = (*(uint16_t*)(msg.buf + 4));
        fuelalarmflags = (*(uint16_t*)(msg.buf + 6));
        lefttank = lefttankraw / 10000.0 * FUELFSLEVEL;
        lefttank = iirFilter(lefttank_Old, lefttank, 0.3);  // filtrage IIR
        lefttank_Old = lefttank;
        righttank = righttankraw / 10000.0 * FUELFSLEVEL;       
        righttank = iirFilter(righttank_Old, righttank, 0.3);  // filtrage IIR
        righttank_Old = righttank;
        break;

      default: break;
    }
    NoFUELmsg = 0;
  }
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                          Fonctionsos de mise à jour des affichages textuels et du vario, et envoi de données au CAN bus
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
void displayTexts() {
  // ------------- Affichage direction et vitesse du vent -------------------
  tft.setCursor(35, 15);
  int16_t windDirectionInt = int(windDirection + 0.5);
  tft.printf("%3d", windDirectionInt);
  tft.setCursor(35, 30);
  int16_t windSpeedInt = int(windSpeed + 0.5);
  tft.printf("%3d", windSpeedInt);
  tft.setCursor(65, 30);
  tft.print(strSpeedUnit);

  // ------------- Affichage IAT, OAT et Dew Point---------------------------
  if (displayTemperatureUnit == 0) {
    tft.setCursor(35, 157);
    tft.printf("%5.1f%s", iat, " C");
    tft.setCursor(35, 172);
    tft.printf("%5.1f%s", oat, " C");
    tft.setCursor(35, 202);
    tft.printf("%5.1f%s", dewPoint, " C");
  } else {
    tft.setCursor(35, 157);
    tft.printf("%5.1f%s", (iat * 1.8 + 32), " F");
    tft.setCursor(35, 172);
    tft.printf("%5.1f%s", (oat * 1.8 + 32), " F");
    tft.setCursor(35, 202);
    tft.printf("%5.1f%s", (dewPoint * 1.8 + 32), " F");
  }

  // --------------Affichage Humidité ---------------------
  tft.setCursor(35, 187);
  tft.printf("%3d%s", int(relativeHumidity + 0.5), " %");

  // ------------- Affichage Vitesse indiquée ---------------------------
  tft.setFontScale(1);
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  tft.setCursor(7, 118);
  tft.printf("%3d", indicatedAirSpeed);  // elle a été calculée dans l'unité choisie par l'utilisateur
  tft.setCursor(65, 127);
  tft.setFontScale(0);
  tft.setTextColor(RA8875_WHITE, RA8875_PINK);
  tft.print(strSpeedUnit);

  // ------------- Affichage Route GPS ---------------------------
  if (groundSpeedGNSS < 5.0) tft.setTextColor(RA8875_WHITE, RA8875_RED);
  else tft.setTextColor(RA8875_WHITE, RA8875_BLACK);  // Si la vitesse GPS est trop faible (moins de 5 km/h), trk n'est pas fiable, et va donc s'afficher sur fond rouge au lieu de noir.
  tft.setFontScale(1, 0);
  tft.setCursor(248, 3);
  tft.printf("%3.0f", trk);

  // --------------Affichage cap magnétique ---------------------
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  tft.setCursor(185, 3);
//  tft.printf("%3d", intMagneticHeading);
  // Enable Heading Correction
  if(magDev >= 0)
    if(intMagneticHeading <= (360 - magDev))
    {
      tft.printf("%3d", intMagneticHeading + magDev);
    }
    else
    {
      tft.printf("%3d", magDev + intMagneticHeading - 360);
    }
  else
    if(intMagneticHeading < -magDev)
    {
      tft.printf("%3d", intMagneticHeading + magDev + 360);
    }
    else
    {
      tft.printf("%3d", intMagneticHeading + magDev);
    }

  // ------------- Affichage Vitesse propre (TAS) ---------------------------
  tft.setFontScale(0);
  tft.setTextColor(RA8875_WHITE, RA8875_PINK);
  tft.setCursor(30, 97);
  tft.printf("%3d", trueAirSpeed);  // elle a été claculée dans l'unité choisie par l'utilisateur
  tft.setCursor(65, 97);
  tft.print(strSpeedUnit);

  // --------------- Affichage vitesse GNSS dans l'unité choisie par l'utilisateur ----------------------------------
  tft.setCursor(30, 82);
  tft.printf("%3.0f", (groundSpeedGNSS / speedConversionFactor) + 0.5);
  tft.setCursor(65, 82);
  tft.print(strSpeedUnit);

  // ------------- Affichage Vz AHRS et Vario barométrique ---------------------------
  tft.setCursor(415, 15);
  tft.printf("%5d", roundedIntVz);
  tft.setCursor(415, 30);
  int16_t baroVarioInt = int(baroVario);
  tft.printf("%5d", baroVarioInt);

  // ------------- Affichage nombre de satellites ---------------------------
  tft.setCursor(3, 48);
  tft.print("Nb SAT : ");
  tft.printf("%2d", int(satInView));

  // ------------- Affichage QNH ---------------------------
  tft.setCursor(415, 205);
  if (displayBaroSettingUnit == 1)  // alors on va afficher le QNH en pouces de mercure
  {
    tft.printf("%02u.%02u", (QNH / 100), (QNH % 100));
  } else  // sinon on va afficher le QNH en hPa
  {
    float i = QNH * 0.338646;
    uint16_t j = int(i + 0.5);
    tft.printf("% 5d", j);
  }

  // ------------- Affichage QFE ---------------------------
  tft.setCursor(415, 220);
  if (displayBaroSettingUnit == 1)  // alors on va afficher le QFE en pouces de mercure
  {
    tft.printf("%02u.%02u", (QFE / 100), (QFE % 100));
  } else  // sinon on va afficher le QFE en hPa
  {
    float i = QFE * 0.338646;
    uint16_t j = int(i + 0.5);
    tft.printf("% 5d", j);
  }

  // ------------- Affichage des 5 altitudes ---------------------------
  tft.setFontScale(0, 1);
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  tft.setCursor(411, 118);
  int16_t qnhAltitudeInt = int(qnhAltitude);
  tft.printf("%5d", qnhAltitudeInt);
  tft.setFontScale(0);
  tft.setCursor(411, 98);
  tft.printf("%5d", int(qfeAltitude));
  tft.setTextColor(RA8875_WHITE, RA8875_PINK);
  tft.setCursor(415, 157);
  tft.printf("%5d", int(pressureAltitude));
  tft.setCursor(415, 172);
  if (densityAltitude > 99999) densityAltitude = 99999;  // pour prévenir un bug d'affichage
  int16_t densityAltitudeInt = int(densityAltitude);
  tft.printf("%5d", densityAltitudeInt);
  tft.setCursor(415, 187);
  uint16_t altGps = int((altitudeGNSS + 5) / 10.0) * 10;
  tft.printf("%5d", altGps);

  // ------------- Affichage G, Gmin et Gmax ---------------------------
  tft.setCursor(378, 63);
  tft.print("G    ");
  tft.printf("%4.1f", accG);
  tft.setCursor(378, 48);
  tft.print("max  ");
  tft.printf("%4.1f", accGmax);
  tft.setCursor(378, 78);
  tft.print("min  ");
  tft.printf("%4.1f", accGmin);

  // ------------- Envois des données au CAN Bus à 5 Hz
  if ((millis() - periode200msStartTime) >= periode200msTimeOut) {
    periode200msStartTime = millis();
    for (uint8_t i = 0; i < 2; i++) {
      msgQNHaltitudesVario.buf[i] = ((byte*)&QNH)[i];
      msgQNHaltitudesVario.buf[i + 2] = ((byte*)&qnhAltitudeInt)[i];
      msgQNHaltitudesVario.buf[i + 4] = ((byte*)&densityAltitudeInt)[i];
      msgQNHaltitudesVario.buf[i + 6] = ((byte*)&baroVarioInt)[i];
      msgSpeeds_Wind.buf[i] = ((byte*)&nonFilteredIndicatedAirSpeed)[i];
      msgSpeeds_Wind.buf[i + 2] = ((byte*)&nonFilteredTrueAirSpeed)[i];
      msgSpeeds_Wind.buf[i + 4] = ((byte*)&windSpeedInt)[i];
      msgSpeeds_Wind.buf[i + 6] = ((byte*)&windDirectionInt)[i];
    }
    CAN_Module_EFIS.write(msgQNHaltitudesVario);
    CAN_Module_EFIS.write(msgSpeeds_Wind);
  }
}

void displayVario() {
  // affichage des barres en premier
  int8_t barHeight;
  if (abs(baroVario) > 2000) barHeight = 116;
  else barHeight = abs(int8_t((baroVario * 58) / 1000));
  if (baroVario < 0) {
    tft.fillRect(461, 20 - 14, 9, 116, RA8875_PINK);
    tft.fillRect(461, 136 - 14, 9, barHeight, RA8875_PURPLE);
    tft.fillRect(461, 136 - 14 + barHeight, 9, 252 - 136 - barHeight, RA8875_PINK);
    tft.drawRect(460, 136 - 14, 10, barHeight + 1, RA8875_WHITE);
  } else {
    tft.fillRect(461, 20 - 14, 9, 116 - barHeight, RA8875_PINK);
    tft.fillRect(461, 136 - 14 - barHeight, 9, barHeight, 0b1111111001000000);
    tft.fillRect(461, 136 - 14, 9, 116, RA8875_PINK);
    tft.drawRect(460, 136 - 14 - barHeight, 10, barHeight + 1, RA8875_WHITE);
  }
  // puis affichage des graduations
  for (uint16_t i = 49; i <= 272; i = i + 58) tft.drawFastHLine(460, i - 14, 6, RA8875_WHITE);
  for (uint16_t i = 20; i <= 272; i = i + 58) tft.drawFastHLine(460, i - 14, 10, RA8875_WHITE);
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                                          Fonctions liées à l'encodeur rotatif et aux menus
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void validateValues()  // Cette fonction vérifie si certains paramètres modifiés dans les menus restent dans les limites autorisées
{
  if (backLightBrightness > 64) backLightBrightness = 64;
  if (backLightBrightness < 1) backLightBrightness = 1;
  if (timerDuration > 60) timerDuration = 60;
  if (timerDuration < 0) timerDuration = 0;
  if (localTimeCorrection < -12) localTimeCorrection = -12;
  if (localTimeCorrection > 12) localTimeCorrection = 12;
}

// ************************************************************************* Routine d'interruption liée au traitement d'une pression sur le bouton *******************************************************************************
void encoderButtonPress_ISR() {
  if ((millis() - lastButtonPress) >= debounceTime) {
    lastButtonPress = millis();
    Button_Click = true;
  }
}
// ***************************************************************************************** Procédure d'ouverture du menu ******************************************************************************************************

void openMenu(uint8_t ndx) {
  indexCurrentMenuOption = ndx;
  tft.fillRect(menuXCoordinate, menuYCoordinate - boxHeight + 2, tft.width() - menuXCoordinate, boxHeight * 2 - 2, BckgrndMenus);
  tft.drawLine(menuXCoordinate + 1, menuYCoordinate - boxHeight + 2, tft.width() - menuXCoordinate, menuYCoordinate - boxHeight + 2, RA8875_WHITE);
  encoder.write(0);
  EncoderPositionValue = 0;
  EncoderPositionValueOld = 0;
  tft.setFont(Arial_11);
  tft.setTextColor(RA8875_WHITE, BckgrndMenus);
  tft.setCursor(menuXCoordinate + 1, menuYCoordinate + 5 - boxHeight);
  if (indexCurrentMenuOption == 0) {
    tft.print("GENERAL MENU");
  } else {
    int8_t i = 0;
    tft.setTextColor(RA8875_WHITE, BckgrndMenus);
    while (menu[indexCurrentMenuOption + i].menuItemNumber != (menu[indexCurrentMenuOption].menuItemNumber) / 10) i--;
    tft.print(menu[indexCurrentMenuOption + i].menuItemLabel);
  }
  for (byte n = ndx; n < ndx + menu[ndx].numberOfSisterOptions; n++) {
    tft.setCursor((boxWidth * (n - ndx)) + 2, menuYCoordinate + 3);
    tft.print(menu[n].menuItemLabel);
  }
  tft.drawRect((EncoderPositionValue * boxWidth), menuYCoordinate, boxWidth, boxHeight, 0xFFFF);
}
// ********************************************************************************************* Procédure de fermeture du menu **************************************************************************************************
void closeMenu() {
  Menu_open = false;
  Button_Click = false;
  tft.fillRect(menuXCoordinate, menuYCoordinate - boxHeight, tft.width() - menuXCoordinate, boxHeight * 2, RA8875_PINK);
  Horizon.reDraw();
  tft.setFontDefault();
  tft.setFontScale(0);
}
// ************************************************ Cette fonction retourne, à partir d'un index donné quelconque, l'index de la première option soeur (numéro terminant par un zéro) **********************************************
uint8_t indexFirstSisterOption(uint8_t arbitraryIndex) {
  int8_t i = 0;
  while (menu[arbitraryIndex + i].menuItemNumber > ((menu[arbitraryIndex].menuItemNumber) / 10) * 10) i--;
  return (arbitraryIndex + i);
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                                Fonctions utilisées lors de l'initialisation des options des menus
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void initTable()  // Cette fonction calcule et initialise les champs "numberOfSisterOptions" et "indexOfParentOption" de chaque enregistrement, lors du setup.
{
  for (uint8_t i = 0; i < numberOfMenuItems(); i++) {
    uint32_t n = menu[i].menuItemNumber;
    int8_t j = 0;
    if ((n / 10) * 10 == n) {
      while (menu[i + j].menuItemNumber < n + 10) j++;
      for (uint8_t k = 0; k < j; k++) menu[i + k].numberOfSisterOptions = j;
      j = 0;
    }
    if (menu[i].menuItemAction == 'S') {
      while (menu[i + j].menuItemNumber != n * 10) j++;
      menu[i].indexOfParentOption = i + j;
      j = 0;
    }
    if (menu[i].menuItemAction == 'B') {
      while (menu[i + j].menuItemNumber != (n / 100) * 10) j--;
      menu[i].indexOfParentOption = i + j;
    }
  }
}

void bubbleSortTable() {
  const char* text;
  uint32_t number;
  char letter;
  for (byte i = numberOfMenuItems() - 1; i > 0; i--) {
    for (byte j = 1; j <= i; j++) {
      if (menu[j - 1].menuItemNumber > menu[j].menuItemNumber) {
        text = menu[j].menuItemLabel;
        number = menu[j].menuItemNumber;
        letter = menu[j].menuItemAction;
        menu[j].menuItemLabel = menu[j - 1].menuItemLabel;
        menu[j].menuItemNumber = menu[j - 1].menuItemNumber;
        menu[j].menuItemAction = menu[j - 1].menuItemAction;
        menu[j - 1].menuItemLabel = text;
        menu[j - 1].menuItemNumber = number;
        menu[j - 1].menuItemAction = letter;
      }
    }
  }
}

uint8_t numberOfMenuItems() {
  uint8_t i = 0;
  while (menu[i].menuItemNumber != 1000000) {
    i = i + 1;
  }
  return i;
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                                Fonction utilisée pour le filtrage des données brutes issues des capteurs
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

// Fonction de filtre à Réponse Impulsionnelle Infinie (RII)
float iirFilter(float previousFilteredValue, float currentValue, float iirFilterCoefficient) {
  return previousFilteredValue * (1 - iirFilterCoefficient) + currentValue * iirFilterCoefficient;
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                                       Fonctions liées à la gestion et à l'affichage de l'heure sur l'écran
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
time_t getTeensy3Time()  // retourne l'heure du RTC dans une variable de type time_t
{
  time_t n;
  n = Teensy3Clock.get();
  n++;  // Correction d'un bug qui fait perdre une seconde sur l'heure RTC
  return n;
}

void displayTime() {
  // affiche l'heure locale digitale sur l'horloge
  tft.setFontDefault();
  tft.setFontScale(0);
  tft.setCursor(Chx - 31, Chy + 30);
  zt = now() + (3600 * localTimeCorrection);  //TimeLib et le RTC sont réglés sur l'heure UTC, donc zt contient maintenant l'heure locale
  tft.printf("%02u:%02u:%02u", hour(zt), minute(zt), second(zt));
  // Calcul des angles des 3 aiguilles
  hs = zt % numberOfSecondsIn12h;
  ms = zt % 3600;
  ss = zt % 60;
  angleH = 360 * hs / numberOfSecondsIn12h;
  angleM = 360 * ms / 3600;
  angleS = 360 * ss / 60;
  // On efface les anciennes aiguilles
  tft.drawLineAngle(Chx, Chy, angleHold, hourHandLength, RA8875_PINK);
  tft.drawLineAngle(Chx, Chy, angleMold, minuteHandLength, RA8875_PINK);
  tft.drawLineAngle(Chx, Chy, angleSold, secondHandLength, RA8875_PINK);
  // On affiche les nouvelles aiguilles
  tft.drawLineAngle(Chx, Chy, angleH, hourHandLength, 0xFFFF);
  tft.drawLineAngle(Chx, Chy, angleM, minuteHandLength, 0xFFFF);
  tft.drawLineAngle(Chx, Chy, angleS, secondHandLength, RA8875_DARK_ORANGE);
  // Sans oublier le petit cercle au centre
  tft.fillCircle(Chx, Chy, 3, RA8875_RED);
  //  Et on mémorise les positions des aiguilles
  angleHold = angleH;
  angleMold = angleM;
  angleSold = angleS;
}


void setflagDisplayTime() {
  flagDisplayTime = true;
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//                                                                                Fonctions utilitaires diverses
//--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

void drawHourGlass(uint8_t x, uint8_t y)  // affiche un sablier
{
  tft.fillRect(x - 20, y - 26, 40, 2, RA8875_WHITE);
  tft.fillRect(x - 20, y + 25, 40, 2, RA8875_WHITE);
  tft.drawLine(x - 18, y - 21, x - 18, y + 20, RA8875_WHITE);
  tft.drawLine(x + 18, y - 21, x + 18, y + 20, RA8875_WHITE);
  tft.drawArc(x, y - 14, 14, 1, 195, 270, RA8875_WHITE);
  tft.drawArc(x, y - 14, 14, 1, 90, 165, RA8875_WHITE);
  tft.drawArc(x, y + 14, 14, 1, 270, 345, RA8875_WHITE);
  tft.drawArc(x, y + 14, 14, 1, 15, 90, RA8875_WHITE);
  tft.drawArc(x, y + 9, 32, 1, 335, 25, RA8875_WHITE);
  tft.drawArc(x, y - 9, 32, 1, 155, 205, RA8875_WHITE);
  tft.drawLine(x - 13, y - 14, x - 13, y - 19, RA8875_WHITE);
  tft.drawLine(x - 13, y + 14, x - 13, y + 19, RA8875_WHITE);
  tft.drawLine(x + 13, y - 14, x + 13, y - 19, RA8875_WHITE);
  tft.drawLine(x + 13, y + 14, x + 13, y + 19, RA8875_WHITE);
  tft.drawArc(x, y - 14, 12, 13, 90, 270, RA8875_WHITE);
  tft.drawLine(x - 1, y - 14, x + 1, y - 14, RA8875_WHITE);
  tft.fillRect(x - 1, y - 3, 3, 12, RA8875_WHITE);
  tft.fillTriangle(x, y + 7, x + 8, y + 20, x - 8, y + 20, RA8875_WHITE);
}

void displayPhoto(char str[], uint16_t photoWidth, uint16_t photoHeight)
{
  file = SD.open(str, FILE_READ);
  uint16_t xOrig, yOrig;
  xOrig = (tft.width() - photoWidth) / 2;
  yOrig = (tft.height() - photoHeight) / 2;
  if (file) {
    unsigned char LSB, MSB;
    uint16_t couleur;
    for (uint16_t j = 1; j <= photoHeight; j++) {
      for (uint16_t i = 1; i <= photoWidth; i++) {
        LSB = file.read();
        MSB = file.read();
        couleur = LSB + (MSB << 8);
        tft.drawPixel(i + xOrig, j + yOrig, couleur);
      }
    }
    file.close();
  } else {
    tft.print("Erreur a l'ouverture du fichier ");
    tft.println(str);
    delay(2000);
  }
}

/******************************************* Conversion algorithms *****************************************/
// converts resistance of VDO NTC to temperature [°C]
float VDOr2t(float r)
{
  float t;
  // STEINHART–HART equation for NTC P/N S-WT-VDO150-M10 (Rotax)
  t = 1.572336E-03 + (2.732929E-04 * log(r)) + (-2.862122E-07 * pow(log(r), 3));
  t = (1 / t) - 273.15;
  return t;
}

// Linear interpolation function for PT1000
float interpolate(float x, float xTable[], float yTable[], int size)
{
  for (int i = 1; i < size; i++)
  {
    if (x <= xTable[i])
    {
      // Linear interpolation
      float x0 = xTable[i - 1];
      float x1 = xTable[i];
      float y0 = yTable[i - 1];
      float y1 = yTable[i];
      return y0 + ((y1 - y0) * (x - x0) / (x1 - x0));
    }
  }
  // Extrapolation if x is outside the range of the table
  return yTable[size - 1];
}

// Linear model for PT1000 Platinum Sender
// C. Marcheggiani 2019.06.28
double PTSender(double x_in)
{
  const double alpha = 0.00385;

  // Algoritm protection
  if (x_in < 800) {  // -50 C
    return -99;
  }

  if (x_in > 1750) {  // +200 C
    return 999;
  }

  return ((x_in / 1000) - 1) / alpha;
}

/************************************* Graph methods *****************************************/
void DrawRpmMap()
{
  tft.setFontScale(2);
  // RPM
  if (rpm > 9999)
    rpm = 9999;

  if (rpm < rpmold - 9 || rpm > rpmold + 9 || rpmold == 0) {
    rpmold = rpm;
    if (rpm >= RPMMAX) {
      tft.setTextColor(RA8875_WHITE, RA8875_RED);
    } else if (rpm > RPMNOR) {
      tft.setTextColor(RA8875_BLACK, RA8875_YELLOW);
    } else {
      tft.setTextColor(RA8875_GREEN, RA8875_BLACK);
    }
    tft.setCursor(10, 280);
    tft.printf("%04u", rpm / 10 * 10);
  }

  // MAP
  if (Map < mapold - 0.5 || Map > mapold + 0.5) {
    mapold = Map;
    if (Map >= MAPMAX) {
      tft.setTextColor(RA8875_WHITE, RA8875_RED);
    } else {
      tft.setTextColor(RA8875_GREEN, RA8875_BLACK);
    }
    tft.setCursor(140, 280);
    tft.printf("%2.1f", Map);
  }
  tft.setFontScale(0);
}

void DrawCAT()
{
  // CAT
  tft.setFontScale(1);
  if (cat < CATMIN || cat > CATMAX) {
    tft.setTextColor(RA8875_WHITE, RA8875_RED);
  } else if (cat == 0)  // Mettere qui calcolo con DP o RH
  {
    tft.setTextColor(RA8875_BLACK, RA8875_YELLOW);
  } else {
    tft.setTextColor(RA8875_CYAN, RA8875_BLACK);
  }
  tft.setCursor(10, 355);
  tft.printf("%2.1f", cat);
  tft.setFontScale(0);
}

void DrawFlapGear()
{
  tft.setFontScale(1);
  if(NoRELAYflag == false)
  {
    // FLAP
    if (flapstatus != flaprequest)
      tft.setTextColor(RA8875_WHITE, RA8875_LIGHT_ORANGE);
    else if (flapstatus == DOWN)
      tft.setTextColor(RA8875_BLACK, RA8875_GREEN);
    else
      tft.setTextColor(RA8875_LIGHT_GREY, RA8875_BLACK);

    tft.setCursor(100, 355);
    if (flapstatus == DOWN)
      tft.print("DWN");
    else
      tft.print("UP ");

    // Landing Gear
    if (gearstatus != gearrequest)
      tft.setTextColor(RA8875_WHITE, RA8875_RED);
    else if (gearstatus == DOWN)
      tft.setTextColor(RA8875_BLACK, RA8875_GREEN);
    else
      tft.setTextColor(RA8875_YELLOW, RA8875_BLACK);

    tft.setCursor(180, 355);
    if (gearstatus == DOWN)
      tft.print("DWN");
    else
      tft.print("UP ");

  }
  tft.setFontScale(0);
}
void DrawEgt()
{
  uint8_t i = 0;
  uint16_t color = RA8875_GREEN;
  int16_t eg = 0;
  static int16_t egmax[4];
  float egtmin = 1000;
  float egtmax = -100;

  for (i = 0; i < 4; i++) {
    if (egt[i] < egtmin)
      egtmin = egt[i];
    if (egt[i] > egtmax)
      egtmax = egt[i];

    if (egt[i] > EGTMAX)
      color = RA8875_RED;
    else if (egt[i] > EGTNOR && egt[i] <= EGTMAX)
      color = RA8875_YELLOW;
    else
      color = RA8875_GREEN;

    eg = (int16_t)round((egt[i] - 500) / 5);

    if (eg > 100)
      eg = 100;
    if (eg < 2)
      eg = 2;

    if(eg > egmax[i])
      egmax[i] = eg;

    tft.fillRect(10 + (i * 35), 440, 20, 100 - eg, RA8875_BLACK);
    tft.fillRect(10 + (i * 35), 440 + 100 - eg, 20, eg, color);

    if(egmax[i]>45)
      tft.fillRect(10 + (i * 35), 440 + 100 - egmax[i]-2, 20, 2, RA8875_MAGENTA);
  }
  tft.setFontScale(0);
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  tft.setCursor(5, 550);
  tft.printf("Lo:% 3u Hi:% 3u", (int16_t)round(egtmin), (int16_t)round(egtmax));
  if (egtmax - egtmin > EGTDELTA)
    tft.setTextColor(RA8875_LIGHT_ORANGE, RA8875_BLACK);

  tft.setCursor(5, 570);
  tft.printf("EGT dT:% 3u\xB0\x43", (uint16_t)round(egtmax - egtmin));
}

void DrawCht()
{
  uint8_t i = 0;
  uint16_t color = RA8875_GREEN;
  int16_t ch = 0;
  static int16_t chmax[2];
  float chtmin = 200;
  float chtmax = -100;

  for (i = 0; i < 2; i++) {
    if (cht[i] < chtmin)
      chtmin = cht[i];
    if (cht[i] > chtmax)
      chtmax = cht[i];

    if (cht[i] > CHTMAX)
      color = RA8875_RED;
    else if (cht[i] > CHTNOR && cht[i] <= CHTMAX)
      color = RA8875_YELLOW;
    else if (cht[i] >= CHTMIN && cht[i] <= CHTNOR)
      color = RA8875_GREEN;
    else
      color = RA8875_YELLOW;

    ch = (int16_t)round(cht[i] - 50);

    if (ch > 100)
      ch = 100;
    if (ch < 2)
      ch = 2;

    if(ch > chmax[i])
      chmax[i] = ch;

    tft.fillRect(175 + (i * 35), 440, 20, 100 - ch, RA8875_BLACK);
    tft.fillRect(175 + (i * 35), 440 + 100 - ch, 20, ch, color);

    if(chmax[i]>50)
      tft.fillRect(175 + (i * 35), 440 + 100 - chmax[i]-2, 20, 2, RA8875_MAGENTA);
  }
  tft.setFontScale(0);
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  tft.setCursor(125, 550);
  tft.printf("Lo:%-3d Hi:%-3d", (int16_t)round(chtmin), (int16_t)round(chtmax));
  if (chtmax - chtmin > CHTDELTA)
    tft.setTextColor(RA8875_LIGHT_ORANGE, RA8875_BLACK);

  tft.setCursor(125, 570);
  tft.printf("CHT dT:%-2d\xB0\x43", (uint16_t)round(chtmax - chtmin));
}

void DrawOilP()
{
  float p = oilpress;
  uint16_t w = 0;

  if (oilpress < -0.2)
  {
    p = 0.0;
    tft.drawRect(249, 283, 220, 20, RA8875_LIGHT_ORANGE);
    tft.setTextColor(RA8875_LIGHT_ORANGE, RA8875_BLACK);
    tft.setCursor(260, 310);
    tft.print("OIL PRESSURE SENSOR ERROR");
//    tft.printf("%f", (float)(curraw*VREF/ADCMAX));
  }
  else
  {
    if (oilpress <= 0)
      p = 0.15;
    else if (oilpress > OILPFS)
      p = OILPFS;

    w = (uint16_t)round((p / OILPFS) * 218);
    if (p < OILPMIN || p > OILPMAX)
      tft.fillRect(250, 284, w, 18, RA8875_RED);
    else if (p > OILPNOR)
      tft.fillRect(250, 284, w, 18, RA8875_YELLOW);
    else
      tft.fillRect(250, 284, w, 18, RA8875_GREEN);

    tft.fillRect(250 + w + 1, 284, 220 - w - 3, 18, RA8875_BLACK);
    tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
    tft.setCursor(260, 310);
    tft.printf("OIL PRESSURE %0.1f Bar", oilpress);
  }
}

void DrawOilT()
{
  float t = oiltemp;
  uint16_t w = 0;

  if (oiltemp < -20)
  {
    t = 0.0;
    tft.drawRect(249, 340, 220, 20, RA8875_LIGHT_ORANGE);
    tft.setTextColor(RA8875_LIGHT_ORANGE, RA8875_BLACK);
    tft.setCursor(250, 367);
    tft.print("OIL TEMPERATURE SENSOR ERROR");
  }
  else 
  {
    if (oiltemp <= OILTZER)
      t = OILTZER + 2;
    else if (oiltemp > OILTFS)
      t = OILTFS;

    w = (uint16_t)round(((t - OILTZER) / (OILTFS - OILTZER)) * 218);
    if (t > OILTMAX)
      tft.fillRect(250, 341, w, 18, RA8875_RED);
    else if (t < OILTMIN || t > OILTNOR)
      tft.fillRect(250, 341, w, 18, RA8875_YELLOW);
    else
      tft.fillRect(250, 341, w, 18, RA8875_GREEN);

    tft.fillRect(250 + w + 1, 341, 220 - w - 3, 18, RA8875_BLACK);
    tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
    tft.setCursor(260, 367);
    tft.printf("OIL TEMPERATURE %3d\xB0\x43", (int16_t)round(oiltemp));
  }
}

void DrawTanks()
{
  uint8_t l = (uint8_t)round(lefttank);
  uint8_t r = (uint8_t)round(righttank);
  static uint8_t l_old;
  static uint8_t r_old;

  // update the fuel tanks display only if something is changed
  if (abs(l - l_old) >= FUEL_UPDATE_LEVEL || abs(r - r_old) >= FUEL_UPDATE_LEVEL || fuelalarmflags != 0 || NoFUELmsg > FUEL_MSGS_TIMEOUT) {
    uint16_t h = 0;

    l_old = l;
    r_old = r;

    if (fuelalarmflags != 0 || NoFUELmsg > FUEL_MSGS_TIMEOUT)
    {
      l = 0;
      r = 0;
      lefttank = 0;
      righttank = 0;
    }

    if (l <= FUELZERLEVEL)
      l = FUELZERLEVEL + 1;
    if (l > FUELFSLEVEL)
      l = FUELFSLEVEL;
    h = 98 * l / FUELFSLEVEL;

    tft.fillRect(250, 397, 78, 98 - h - 1, RA8875_BLACK);
    if (lefttank <= FUELMINLEVEL)
      tft.fillRect(250, 397 + 98 - h, 78, h, RA8875_RED);
    else if (lefttank < FUELNORLEVEL)
      tft.fillRect(250, 397 + 98 - h, 78, h, RA8875_YELLOW);
    else
      tft.fillRect(250, 397 + 98 - h, 78, h, RA8875_GREEN);

    if (r <= FUELZERLEVEL)
      r = FUELZERLEVEL + 1;
    if (r > FUELFSLEVEL)
      r = FUELFSLEVEL;
    h = 98 * r / FUELFSLEVEL;

    tft.fillRect(389, 397, 78, 98 - h - 1, RA8875_BLACK);
    if (righttank <= FUELMINLEVEL)
      tft.fillRect(389, 397 + 98 - h, 78, h, RA8875_RED);
    else if (righttank < FUELNORLEVEL)
      tft.fillRect(389, 397 + 98 - h, 78, h, RA8875_YELLOW);
    else
      tft.fillRect(389, 397 + 98 - h, 78, h, RA8875_GREEN);

    tft.fillRect(280, 500, 199, 20, RA8875_BLACK);

    if (fuelalarmflags != 0 || NoFUELmsg > FUEL_MSGS_TIMEOUT)
    {
      tft.setTextColor(RA8875_WHITE, RA8875_RED);
      tft.setCursor(280, 500);
      tft.print("MISSING FUEL MESSAGE");
      tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
    }
    else
    {
      tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
      tft.setCursor(280, 500);
      tft.printf("%2d     LITERS    %2d", (uint8_t)lefttank, (uint8_t)righttank);
    }
  }
}

void DrawFuelP()
{
  float p = fuelpress;
  uint16_t w = 0;

  if (fuelpress < -0.1)
  {
    p = 0.0;
    tft.drawRect(249, 530, 220, 20, RA8875_LIGHT_ORANGE);
    tft.setTextColor(RA8875_LIGHT_ORANGE, RA8875_BLACK);
    tft.setCursor(255, 557);
    tft.print("FUEL PRESSURE SENSOR ERROR");
  } else {
    if (fuelpress < FPZER)
      p = 0.01;
    else if (fuelpress > FPFS)
      p = FPFS;

    w = (uint16_t)((p / FPFS) * 218);
    if (p < FPMIN || p > FPMAX)
      tft.fillRect(250, 531, w, 18, RA8875_RED);
    else
      tft.fillRect(250, 531, w, 18, RA8875_GREEN);

    tft.fillRect(250 + w + 1, 531, 220 - w - 3, 18, RA8875_BLACK);
    tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
    tft.setCursor(265, 557);
    tft.printf("FUEL PRESSURE %0.2f Bar", fuelpress);
  }
}

void DrawFuelComp()
{
  if (fuelcomputerflag == true) {
    tft.setCursor(265, 610);
    tft.printf("AVAILABLE FUEL %2d LITERS", availablefuel);
    tft.setCursor(265, 630);
    tft.printf("BURNT FUEL %2d LITERS", fuelburnt);
    if (fuelflowratio == 65535)
    {
      tft.setTextColor(RA8875_LIGHT_ORANGE, RA8875_BLACK);
      tft.setCursor(265, 650);
      tft.print("FUEL FLOW: NO SIGNAL");
    }
    else
    {
      tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
      tft.setCursor(265, 650);
      tft.printf("FUEL FLOW %2.1f LPH   ", fuelflow);
    }
    tft.setCursor(265, 670);
    if (fuelflow > 0 && fuelflowratio != 65535)
    {
      tft.setCursor(265, 670);
      tft.print("                         ");
      tft.setCursor(265, 670);
      tft.printf("ENDURANCE %2.1f HOURS", (float)(availablefuel / fuelflow));
    }
    else
      tft.print("ENDURANCE: UNDETERMINATED");
    fuelcomputerflag = false;
  }
}

void DrawElectric()
{
  bool almgen = false;
  // Generator
  if ((rpm < 500 && volt < VLTMIN) || (rpm > 500 && volt < VLTMIN1) || volt > VLTMAX)
  {
    almgen = true;
    tft.fillCircle(40, 664, 20, RA8875_RED);
  }
  else if (volt > VLTCHG)
  {
    almgen = false;
    tft.fillCircle(40, 664, 20, RA8875_CYAN);
  }
  else
  {
    almgen = false;
    tft.fillCircle(40, 664, 20, RA8875_GREEN);
  }

  tft.fillCircle(40, 664, 18, RA8875_BLACK);
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  tft.setCursor(36, 655);
  tft.print("G");
  // Electrical values
  if (almgen)
    tft.setTextColor(RA8875_WHITE, RA8875_RED);
  else
    tft.setTextColor(RA8875_WHITE, RA8875_BLACK);

  tft.setCursor(125, 595);
  tft.printf(" %2.1fV ", volt);
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  tft.setCursor(80, 620);
  tft.printf("% 2.1fA      % 2.1fA", ampbat, ampbus);
  tft.setCursor(40, 740);
  tft.printf("GEN. POWER %2.1fW ", volt * (ampbat + ampbus));
  if ((-volt * ampbat) > 0)
    tft.setTextColor(RA8875_BLACK, RA8875_YELLOW);
  else
    tft.setTextColor(RA8875_WHITE, RA8875_BLACK);

  tft.setCursor(40, 760);
  tft.printf("BAT. POWER %2.1fW ", -volt * ampbat);
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
}

void DrawEMSframe()
{
  tft.fillRect(0, 273, 479, 799, RA8875_BLACK);
  tft.drawRect(0, 273, 479, 799, RA8875_DARK_GREEN);
  tft.drawFastVLine(239, 273, 526, RA8875_DARK_GREEN);
  tft.drawFastHLine(0, 350, 240, RA8875_DARK_GREEN);
  tft.drawFastHLine(0, 410, 240, RA8875_DARK_GREEN);
  tft.setTextColor(RA8875_LIGHT_GREY, RA8875_BLACK);
  tft.setCursor(45, 330);
  tft.print("RPM         MAP HgInch");
  tft.setCursor(20, 390);
  tft.print("CAT\xB0\x43    FLAP   LAND.GEAR");
  tft.setTextColor(RA8875_BLUE, RA8875_BLACK);
  tft.setCursor(20, 415);
  tft.print("1   2   3   4       1   2");
  tft.fillRect(5, 540, 229, 3, RA8875_LIGHT_GREY);
  tft.drawFastHLine(5, 539 - ((EGTNOR - 500) / 5), 134, RA8875_YELLOW);
  tft.drawFastHLine(5, 539 - ((EGTMAX - 500) / 5), 134, RA8875_RED);
  tft.drawFastHLine(170, 539 - (CHTMIN - 50), 65, RA8875_GREEN);
  tft.drawFastHLine(170, 539 - (CHTNOR - 50), 65, RA8875_YELLOW);
  tft.drawFastHLine(170, 539 - (CHTMAX - 50), 65, RA8875_RED);
  tft.drawFastHLine(0, 590, 240, RA8875_DARK_GREEN);
  // oilp
  tft.drawRect(249, 283, 220, 20, RA8875_LIGHT_GREY);
  tft.drawFastHLine(239, 330, 240, RA8875_DARK_GREEN);
  // oilt
  tft.drawRect(249, 340, 220, 20, RA8875_LIGHT_GREY);
  tft.drawFastHLine(239, 387, 240, RA8875_DARK_GREEN);
  // Fuel tanks
  tft.drawRect(249, 396, 80, 100, RA8875_LIGHT_GREY);
  tft.drawRect(388, 396, 80, 100, RA8875_LIGHT_GREY);
  const char f[2][7] = { "L TANK", "R TANK" };
  for (uint8_t i = 0; i < 6; i++)
  {
    tft.setCursor(335, 396 + (i * 15));
    tft.printf("%c    %c", f[0][i], f[1][i]);
  }
  tft.drawFastHLine(239, 520, 240, RA8875_DARK_GREEN);
  // FuelP
  tft.drawRect(249, 530, 220, 20, RA8875_LIGHT_GREY);
  tft.drawFastHLine(239, 577, 240, RA8875_DARK_GREEN);
  // Fuel Computer
  tft.setTextColor(RA8875_BLUE, RA8875_BLACK);
  tft.setCursor(305, 585);
  tft.print("FUEL COMPUTER");
  tft.drawFastHLine(239, 690, 240, RA8875_DARK_GREEN);
  // Electrical plant
  // Wires
  tft.drawRect(40, 605, 155, 120, RA8875_WHITE);
  tft.drawRect(127, 637, 20, 88, RA8875_WHITE);
  tft.drawFastVLine(127, 682, 42, RA8875_BLACK);
  tft.drawFastVLine(100, 605, 42, RA8875_WHITE);
  tft.fillCircle(100, 605, 2, RA8875_WHITE);
  tft.fillCircle(146, 724, 2, RA8875_WHITE);
  // Generator
  if ((rpm < 500 && volt < VLTMIN) || (rpm > 500 && volt < VLTMIN1) || volt > VLTMAX)
    tft.fillCircle(40, 664, 20, RA8875_RED);
  else if (volt > VLTCHG)
    tft.fillCircle(40, 664, 20, RA8875_CYAN);
  else
    tft.fillCircle(40, 664, 20, RA8875_GREEN);
  tft.fillCircle(40, 664, 18, RA8875_BLACK);
  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
  tft.setCursor(36, 655);
  tft.print("G");
  // Battery
  tft.drawRect(90, 647, 50, 35, RA8875_LIGHT_GREY);
  tft.fillRect(91, 648, 48, 33, RA8875_BLACK);
  tft.setTextColor(RA8875_LIGHT_GREY, RA8875_BLACK);
  tft.setCursor(95, 648);
  tft.print("+   -");
  // Aircraft load
  tft.drawRect(170, 647, 50, 35, RA8875_LIGHT_GREY);
  tft.fillRect(171, 648, 48, 33, RA8875_BLACK);
  tft.setTextColor(RA8875_LIGHT_GREY, RA8875_BLACK);
  tft.setCursor(179, 655);
  tft.print("LOAD");

  tft.setTextColor(RA8875_BLUE, RA8875_BLACK);
  tft.setCursor(315, 725);
  tft.print("            ");

  tft.setTextColor(RA8875_WHITE, RA8875_BLACK);
}

void DrawMessage()
{
  switch (alm.errclass)
  {
    case ALARM:
      tft.setTextColor(RA8875_RED, RA8875_BLACK);
      tft.setCursor(315, 725);
      tft.print(" DANGER: ");
    break;

    case WARNING:
      tft.setTextColor(RA8875_YELLOW, RA8875_BLACK);
      tft.setCursor(315, 725);
      tft.print("WARNING: ");
    break;

    case ADVISOR:
      tft.setTextColor(RA8875_GREEN, RA8875_BLACK);
      tft.setCursor(315, 725);
      tft.print(" INFO:   ");
    break;

    case NOP:
    default:
      tft.setTextColor(RA8875_BLUE, RA8875_BLACK);
      tft.setCursor(315, 725);
      tft.print("              ");
    break;
  }
  tft.setCursor(275, 745);
  tft.print("                       ");
  tft.setCursor(275, 745);
  tft.print(alm.msg);
}

void ResetVars()
{
  uint8_t i = 0;
  oiltemp = 0.0;
  oiltemp_Old = 0.0;
  oilpress = 0.0;
  oilpress_Old = 0.0;
  fuelpress = 0.0;
  fuelpress_Old = 0.0;
  fuelflow = 0;
  fuelflow_Old = 0;
  lefttank = 0.0;
  righttank = 0.0;
  lefttank_Old = 0.0;
  righttank_Old = 0.0;
  for(i=0;i<4;i++)
  {
    egt[i] = 0.0;
    egt_Old[i] = 0.0;
  }
  for(i=0;i<2;i++)
  {
    cht[i] = 0.0;
    cht_Old[i] = 0.0;
  }  
  cat = 0.0;
  cat_Old = 0.0;
  volt = 0.0;
  volt_Old = 0.0;
  ampbat = 0.0;
  ampbus = 0.0;
  ampbat_Old = 0.0;
  ampbus_Old = 0.0;
  Map = 0.0;
  Map_Old = 0.0;
  mapold = 0.0;
  vtemp = 0.0;
  res = 0.0;
  rpmold = 0;
}

void ShowAlarms()
{	
 	if(rpm > 0)	  // Questi allarmi sono attivi solo a motore acceso
 	{

		// WARMUP
	
		if(cht[0]<CHTMIN)
			alarmflag |= ALM_WARMUP;
		else if(cht[1]<CHTMIN)
			alarmflag |= ALM_WARMUP;
		else if(oiltemp<OILTMIN)
			alarmflag |= ALM_WARMUP;
		else
			alarmflag &= ~ALM_WARMUP;

		// CHT Alarm

		alarmflag &= ~ALM_CHT;
	
		if(cht[0]>CHTMAX || cht[1]>CHTMAX)
			alarmflag |= ALM_CHT;

		// EGT Alarm	
		
	 	alarmflag &= ~ALM_EGT;

		if(egt[0]>EGTMAX||egt[1]>EGTMAX||egt[2]>EGTMAX||egt[3]>EGTMAX)
				alarmflag |= ALM_EGT;

		// Oil temperature Alarm	

		if(oiltemp > OILTMAX)
			alarmflag |= ALM_OILT;
		else
			alarmflag &= ~ALM_OILT;

		// Oil pressure Alarm	
		
		if((oilpress>OILPMAX) || (oilpress<OILPMIN))
			alarmflag |= ALM_OILP;
		else
			alarmflag &= ~ALM_OILP;

		// Fuel pressure Alarm	

		if((fuelpress>FPMAX) || (fuelpress<FPMIN))
			alarmflag |= ALM_FP;
		else
			alarmflag &= ~ALM_FP;

		// Fuel flow Alarm	

		if(fuelburnt<FUELMINBURNT)
			alarmflag |= ALM_FF;
		else
			alarmflag &= ~ALM_FF;

	}
	
	// Fuel level Alarm	
#if NOFTANKS == 3 
	if((lefttank < FUELMINLEVEL) && (righttank < FUELMINLEVEL) && (auxtank < FUELMINLEVEL))
#else
	if((lefttank < FUELMINLEVEL) && (righttank < FUELMINLEVEL))
#endif
		alarmflag |= ALM_FUEL;
	else
		alarmflag &= ~ALM_FUEL;

	// Fuel Tank Switch Alarm	

	if(lefttank < righttank-MAXFUELDIFF)
		alarmflag |= ALM_SWTORTANK;
	else
		alarmflag &= ~ALM_SWTORTANK;
		
	if(righttank < lefttank-MAXFUELDIFF)
		alarmflag |= ALM_SWTOLTANK;
	else
		alarmflag &= ~ALM_SWTOLTANK;	

	// RPM Alarm	
	
	if(rpm > RPMMAX)
		alarmflag |= ALM_RPM;
	else
		alarmflag &= ~ALM_RPM;

	// Map Alarm	
	
	if(Map > MAPMAX)
		alarmflag |= ALM_MAP;
	else
		alarmflag &= ~ALM_MAP;

	// Cat Alarm
	// Carburattor icing protection as EASA EGAST_GA5 raccomandation
	if((cat > CATMAX) || (cat < CATMIN) || ((cat > -2.0 && cat < 20.0) && (dewPoint > -2.0 && dewPoint < 20.0))) 
		alarmflag |= ALM_CAT;
	else
		alarmflag &= ~ALM_CAT;

	// Oat Alarm	
	
	if((oat>OATMAX) || (oat<OATMIN))
		alarmflag |= ALM_OAT;
	else
		alarmflag &= ~ALM_OAT;

	// Volt Alarm	
	
	if((volt>VLTMAX) || (volt<VLTMIN && rpm<2500) || (volt<VLTMIN1 && rpm>=2500)) // Charging alarm with engine running
		alarmflag |= ALM_VOLT;
	else
		alarmflag &= ~ALM_VOLT;

	// Amp Alarm	
	
	if(ampbus > AMPMAX || ampbus < AMPMIN || ampbat > AMPMAX || ampbat < AMPMIN)
		alarmflag |= ALM_AMP;
	else
		alarmflag &= ~ALM_AMP;
		
	// Allarme sbandometro
	if(ballOffset < -BALLMAX)
		alarmflag |= ALM_LFOOT;
	else
		alarmflag &= ~ALM_LFOOT;

	if(ballOffset > BALLMAX )
		alarmflag |= ALM_RFOOT;
	else
		alarmflag &= ~ALM_RFOOT;

 	// AirSpeed Alarm	
	if(indicatedAirSpeed<=VS*1.2 && indicatedAirSpeed>40)	// 120% of VS & >40 Km/h
		alarmflag |= ALM_LOASI;
	else
		alarmflag &= ~ALM_LOASI;
		
	if(indicatedAirSpeed>VNE)	// more than VNE
		alarmflag |= ALM_HIASI;
	else
		alarmflag &= ~ALM_HIASI;

  // Gear Alarm
  if(indicatedAirSpeed<VS*1.2 && gearstatus == UP)
    alarmflag |= ALM_GEAR;
  else
    alarmflag &= ~ALM_GEAR;

  // Flap Advisor
  // Remind to extract flap before takeoff and entering in circuit
  if(flaprequest == UP && indicatedAirSpeed < VF && (indicatedAirSpeed > 30 || rpm > 3700))
  {	
    flaprequest = DOWN;
  }  else if(flaprequest == DOWN && indicatedAirSpeed > VF+10) {		// Remind to retract flap and check during climb
    flaprequest = UP;
  }

  if(flaprequest != flapstatus) {
    alarmflag |= ALM_FLAP;
  } else {
    alarmflag &= ~ALM_FLAP;
  }		

// *****************************************************
// * ALARM DISABLE: Following lines should be remarked *
// *****************************************************

	alarmflag &= ~ALM_FF;	// Fuelflow alarm disabled

// ****************************************************

	// Alarm priority check
	if((alarmflag & ~checked)&ALM_GEAR) {
		alm.priority = ALM_GEAR;
		strcpy(alm.msg,"  GEAR NOT DOWN");
		alm.errclass = ALARM;
		alm.number=1;
	} else if((alarmflag & ~checked)&ALM_LOASI) {
		alm.priority = ALM_LOASI;
		strcpy(alm.msg,"INCREASE AIRSPEED");
		alm.errclass = ALARM;
		alm.number=2;
	} else if((alarmflag & ~checked)&ALM_FLAP) {
		alm.priority = ALM_FLAP;
		strcpy(alm.msg," CONFIGURE FLAPS");
		alm.errclass = WARNING;
		alm.number=3;
	} else if((alarmflag & ~checked)&ALM_HIASI) {
		alm.priority = ALM_HIASI;
		strcpy(alm.msg," REDUCE AIRSPEED");
		alm.errclass = ALARM;
		alm.number=4;
	} else if((alarmflag & ~checked)&ALM_RPM) {
		alm.priority = ALM_RPM;
		strcpy(alm.msg,"  RPM TOO HIGH");
		alm.errclass = ALARM;
		alm.number=5;
	} else if((alarmflag & ~checked)&ALM_OILP) {
		alm.priority = ALM_OILP;
		strcpy(alm.msg,"CHECK OIL PRESSURE");
		alm.errclass = ALARM;
		alm.number=6;
	} else if((alarmflag & ~checked)&ALM_FP) {
		alm.priority = ALM_FP;
		strcpy(alm.msg,"CHECK FUEL PRESSURE");
		alm.errclass = ALARM;
		alm.number=7;
	} else if((alarmflag & ~checked)&ALM_OILT) {
		alm.priority = ALM_OILT;
		strcpy(alm.msg,"CHECK OIL TEMPERATURE");
		alm.errclass = ALARM;
		alm.number=8;
	} else if((alarmflag & ~checked)&ALM_FF) {
		alm.priority = ALM_FF;
		strcpy(alm.msg,"FUEL FLOW TOO LOW");
		alm.errclass = ALARM;
		alm.number=9;
	} else if((alarmflag & ~checked)&ALM_CHT) {
		alm.priority = ALM_CHT;
		strcpy(alm.msg,"CYLINDER HEAD TEMP.");
		alm.errclass = ALARM;
		alm.number=10;
	} else if((alarmflag & ~checked)&ALM_EGT) {
		alm.priority = ALM_EGT;
		strcpy(alm.msg,"EXAUST GAS TEMPERATURE");
		alm.errclass = ALARM;
		alm.number=11;
	} else if((alarmflag & ~checked)&ALM_VOLT) {
		alm.priority = ALM_VOLT;
		strcpy(alm.msg,"VOLTAGE OUT OF LIMITS");
		alm.errclass = WARNING;
		alm.number=12;
	} else if((alarmflag & ~checked)&ALM_AMP) {
		alm.priority = ALM_AMP;
		strcpy(alm.msg,"CURRENT OUT OF LIMITS");
		alm.errclass = WARNING;
		alm.number=13;
	} else if((alarmflag & ~checked)&ALM_CAT) {
		alm.priority = ALM_CAT;
		strcpy(alm.msg,"CARB. AIR TEMPERATURE");
		alm.errclass = WARNING;
		alm.number=14;
	} else if((alarmflag & ~checked)&ALM_FUEL) {
		alm.priority = ALM_FUEL;
		strcpy(alm.msg,"LOW FUEL TANK LEVEL");
		alm.errclass = WARNING;
		alm.number=15;
	} else if((alarmflag & ~checked)&ALM_MAP) {
		alm.priority = ALM_MAP;
		strcpy(alm.msg,"MAP TOO HIGH");
		alm.errclass = WARNING;
		alm.number=16;
	} else if((alarmflag & ~checked)&ALM_LFOOT) {
		alm.priority = ALM_LFOOT;
		strcpy(alm.msg,"LEFT FOOT");
		alm.errclass = WARNING;
		alm.number=17;
	} else if((alarmflag & ~checked)&ALM_RFOOT) {
		alm.priority = ALM_RFOOT;
		strcpy(alm.msg,"RIGHT FOOT");
		alm.errclass = WARNING;
		alm.number=18;
	} else if((alarmflag & ~checked)&ALM_WARMUP) {
		alm.priority = ALM_WARMUP;
		strcpy(alm.msg,"ENGINE WARM UP");
		alm.errclass = WARNING;
		alm.number=19;
	} else if((alarmflag & ~checked)&ALM_OAT) {
		alm.priority = ALM_OAT;
		strcpy(alm.msg,"OUT AIR TEMPERATURE");
		alm.errclass = ADVISOR;
		alm.number=20;
	} else if((alarmflag & ~checked)&ALM_SWTOLTANK) {
		alm.priority = ALM_SWTOLTANK;
		strcpy(alm.msg,"SWITCH TO LEFT TANK");
		alm.errclass = ADVISOR;
		alm.number=21;
	} else if((alarmflag & ~checked)&ALM_SWTORTANK) {
		alm.priority = ALM_SWTORTANK;
		strcpy(alm.msg,"SWITCH TO RIGHT TANK");
		alm.errclass = ADVISOR;
		alm.number=22;
	} else {
    alm.priority = 0;
		strcpy(alm.msg,"NO MESSAGES");
    alm.errclass = NOP;
		alm.number=0;
	}
	
	if(alarmflag & ALM_SWTOLTANK) {
		checked &= ~ALM_SWTORTANK;
	}

	if(alarmflag & ALM_SWTORTANK) {
		checked &= ~ALM_SWTOLTANK;
	}
}